import os
import argparse
import sys

root_dir = os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))
sys.path.append(root_dir)

from stanza.pipeline.core import Pipeline
from stanza.server.semgrex import Semgrex
from stanza.models.common.constant import is_right_to_left
import spacy
from spacy import displacy
from spacy.tokens import Doc
from IPython.display import display, HTML
import typing
from typing import List, Tuple, Any

from stanza.utils.visualization.utils import find_nth, round_base


def get_sentences_html(doc: Any, language: str, visualize_xpos: bool = False) -> List[str]:
    """
    Returns a list of HTML strings representing the dependency visualizations of a given stanza document.
    One HTML string is generated per sentence of the document object. Converts the stanza document object
    to a spaCy doc object and generates HTML with displaCy.

    @param doc: a stanza document object which can be generated with an NLP pipeline.
    @param language: the two letter language code for the document e.g. "en" for English.
    @param visualize_xpos: A toggled option to use xpos tags for part-of-speech labels instead of upos.

    @return: a list of HTML strings which visualize the dependencies of the doc object.
    """
    USE_FINE_GRAINED = False if not visualize_xpos else True
    html_strings, sentences_to_visualize = [], []
    nlp = spacy.blank(
        "en"
    )  # blank model - we don't use any of the model features, just the visualization
    for sentence in doc.sentences:
        words, lemmas, heads, deps, tags = [], [], [], [], []
        if is_right_to_left(
            language
        ):  # order of words displayed is reversed, dependency arcs remain intact
            sentence_len = len(sentence.words)
            for word in reversed(sentence.words):
                words.append(word.text)
                lemmas.append(word.lemma)
                deps.append(word.deprel)
                if visualize_xpos and word.xpos:
                    tags.append(word.xpos)
                else:
                    tags.append(word.upos)
                if word.head == 0:  # spaCy head indexes are one-off from Stanza's
                    heads.append(sentence_len - word.id)
                else:
                    heads.append(sentence_len - word.head)
        else:  # left to right rendering
            for word in sentence.words:
                words.append(word.text)
                lemmas.append(word.lemma)
                deps.append(word.deprel)
                if visualize_xpos and word.xpos:
                    tags.append(word.xpos)
                else:
                    tags.append(word.upos)
                if word.head == 0:
                    heads.append(word.id - 1)
                else:
                    heads.append(word.head - 1)
        if USE_FINE_GRAINED:
            stanza_to_spacy_doc = Doc(
                nlp.vocab, words=words, lemmas=lemmas, heads=heads, deps=deps, tags=tags
            )
        else:
            stanza_to_spacy_doc = Doc(
                nlp.vocab, words=words, lemmas=lemmas, heads=heads, deps=deps, pos=tags
            )
        sentences_to_visualize.append(stanza_to_spacy_doc)

    for line in sentences_to_visualize:  # render all sentences through displaCy
        html_strings.append(
            displacy.render(
                line,
                style="dep",
                options={
                    "compact": True,
                    "word_spacing": 30,
                    "distance": 100,
                    "arrow_spacing": 20,
                    "fine_grained": USE_FINE_GRAINED
                },
                jupyter=False,
            )
        )
    return html_strings


def semgrexify_html(orig_html: str, semgrex_sentence) -> str:
    """
    Modifies the HTML of a sentence's dependency visualization, highlighting words involved in the
    semgrex_sentence search queries and adding the label of the word inside of the match.


    @param orig_html: unedited HTML of a sentence's dependency visualization.
    @param semgrex_sentence: a Semgrex result object containing the matches to a provided query.
    @return: edited HTML containing the visual changes described above.
    """
    tracker = {}  # keep track of which words have multiple labels
    DEFAULT_TSPAN_COUNT = (
        2  # the original displacy html assigns two <tspan> objects per <text> object
    )
    CLOSING_TSPAN_LEN = 8  # </tspan> is 8 chars long
    colors = [
        "#4477AA",
        "#66CCEE",
        "#228833",
        "#CCBB44",
        "#EE6677",
        "#AA3377",
        "#BBBBBB",
    ]  # colorblind-friendly scheme
    css_bolded_class = "<style> .bolded{font-weight: bold;} </style>\n"
    opening_svg_end_idx = orig_html.find("\n")
    # insert the new style class
    orig_html = (
        orig_html[: opening_svg_end_idx + 1]
        + css_bolded_class
        + orig_html[opening_svg_end_idx + 1 :]
    )

    # Color and bold words involved in each Semgrex match
    for query in semgrex_sentence.result:
        for i, match in enumerate(query.match):
            color = colors[i]
            paired_dy = 2
            for node in match.node:
                name, match_index = node.name, node.matchIndex
                # edit existing <tspan> to change color and bold the text
                start = find_nth(
                    orig_html, "<text", match_index
                )  # finds start of svg <text> of interest
                if (
                    match_index not in tracker
                ):  # if we've already bolded and colored, keep the first color
                    tspan_start = orig_html.find(
                        "<tspan", start
                    )  # finds start of the first svg <tspan> inside of the <text>
                    tspan_end = orig_html.find(
                        "</tspan>", start
                    )  # finds start of the end of the above <tspan>
                    tspan_substr = (
                        orig_html[tspan_start : tspan_end + CLOSING_TSPAN_LEN + 1]
                        + "\n"
                    )
                    # color and bold words in the search hit
                    edited_tspan = tspan_substr.replace(
                        'class="displacy-word"', 'class="bolded"'
                    ).replace('fill="currentColor"', f'fill="{color}"')
                    # insert edited <tspan> object into html string

                    # TODO: DEBUG. This code has a bug in it that causes the svg to not end on an input like
                    # "The Wimbledon grass-court tennis tournament banned players, resulting in players hating others."
                    # to malfunction and add another <svg> copy to the tail-end of the first svg rendering.
                    # This bug has been patched in the end of this function, but need to find out what is going on.
                    orig_html = (
                        orig_html[:tspan_start]
                        + edited_tspan
                        + orig_html[tspan_end + CLOSING_TSPAN_LEN + 2 :]
                    )

                    tracker[match_index] = DEFAULT_TSPAN_COUNT

                # next, we have to insert the new <tspan> object for the label
                # Copy old <tspan> to copy formatting when creating new <tspan> later
                prev_tspan_start = (
                    find_nth(orig_html[start:], "<tspan", tracker[match_index] - 1)
                    + start
                )  # find the previous <tspan> start index
                prev_tspan_end = (
                    find_nth(orig_html[start:], "</tspan>", tracker[match_index] - 1)
                    + start
                )  # find the prev </tspan> start index
                prev_tspan = orig_html[
                    prev_tspan_start : prev_tspan_end + CLOSING_TSPAN_LEN + 1
                ]

                # Find spot to insert new tspan
                closing_tspan_start = (
                    find_nth(orig_html[start:], "</tspan>", tracker[match_index])
                    + start
                )
                up_to_new_tspan = orig_html[
                    : closing_tspan_start + CLOSING_TSPAN_LEN + 1
                ]
                rest = orig_html[closing_tspan_start + CLOSING_TSPAN_LEN + 1 :]

                # Calculate proper x value in svg
                x_value_start = prev_tspan.find('x="')
                x_value_end = (
                    prev_tspan[x_value_start + 3 :].find('"') + 3
                )  # 3 is the length of the 'x="' substring
                x_value = prev_tspan[x_value_start + 3 : x_value_end + x_value_start]

                # Calculate proper y value in svg
                DEFAULT_DY_VAL, dy = 2, 2
                if (
                    paired_dy != DEFAULT_DY_VAL and node == match.node[1]
                ):  # we're on the second node and need to adjust height to match the paired node
                    dy = paired_dy
                if node == match.node[0] and len(match.node) > 1:
                    paired_node_level = 2
                    if (
                        match.node[1].matchIndex in tracker
                    ):  # check if we need to adjust heights of labels
                        paired_node_level = tracker[match.node[1].matchIndex]
                        dif = tracker[match_index] - paired_node_level
                        if dif > 0:  # current node has more labels
                            paired_dy = DEFAULT_DY_VAL * dif + 1
                            dy = DEFAULT_DY_VAL
                        else:  # paired node has more labels, adjust this label down
                            dy = DEFAULT_DY_VAL * (abs(dif) + 1)
                            paired_dy = DEFAULT_DY_VAL

                # Insert new <tspan> object
                new_tspan = f'  <tspan class="displacy-word" dy="{dy}em" fill="{color}" x={x_value}>{name[: 3].title()}.</tspan>\n'  # abbreviate label names to 3 chars
                orig_html = up_to_new_tspan + new_tspan + rest
                tracker[match_index] += 1

        # process out extra term if present -- TODO: Figure out why the semgrexify_html function lines 164-168 cause a duplication bug
        end = find_nth(haystack=orig_html, needle="</svg", n=1)
        LENGTH_OF_END_SVG = 7  # </svg> has length 6 so add 1 to the end too
        if len(orig_html) > end + LENGTH_OF_END_SVG:
            orig_html = orig_html[: end + LENGTH_OF_END_SVG]

    return orig_html


def render_html_strings(edited_html_strings: List[str]) -> None:
    """
    Renders the HTML of each HTML string.
    """
    for html_string in edited_html_strings:
        display(HTML(html_string))


def visualize_search_doc(
    doc: Any,
    semgrex_queries: List[str],
    lang_code: str,
    start_match: int = 0,
    end_match: int = 11,
    render: bool = True,
    visualize_xpos: bool = False
) -> List[str]:
    """
    Visualizes the result of running Semgrex search on a document. The i-th element of
    the returned list is the HTML representation of the i-th sentence's dependency
    relationships. Only shows sentences that have a match on the Semgrex search.

    @param doc: A Stanza document object that contains dependency relationships .
    @param semgrex_queries: A list of Semgrex queries to search for in the document.
    @param lang_code: A two letter language abbreviation for the language that the Stanza document is written in.
    @param start_match: Beginning of the splice for which to display elements with.
    @param end_match: End of the splice for which to display elements with.
    @param render: A toggled option to render the HTML strings within the returned list
    @param visualize_xpos: A toggled option to use xpos tags in part-of-speech labels, defaulting to upos tags.

    @return: A list of HTML strings representing the dependency relations of the doc object.
    """

    matches_count = 0  # Limits number of visualizations
    with Semgrex(classpath="$CLASSPATH") as sem:
        edited_html_strings = []
        semgrex_results = sem.process(doc, *semgrex_queries)
        # one html string for each sentence
        unedited_html_strings = get_sentences_html(doc, lang_code, visualize_xpos=visualize_xpos)

        for i in range(len(unedited_html_strings)):

            if matches_count >= end_match:  # we've collected enough matches
                break

            # check if sentence has matches, if not then do not visualize
            has_none = True
            for query in semgrex_results.result[i].result:
                for match in query.match:
                    if match:
                        has_none = False

            # Process HTML if queries have matches
            if not has_none:
                if start_match <= matches_count < end_match:
                    edited_string = semgrexify_html(
                        unedited_html_strings[i], semgrex_results.result[i]
                    )

                    edited_string = adjust_dep_arrows(edited_string)
                    edited_html_strings.append(edited_string)
                matches_count += 1
        if render:
            render_html_strings(edited_html_strings)
    return edited_html_strings


def visualize_search_str(
    text: str,
    semgrex_queries: List[str],
    lang_code: str,
    start_match: int = 0,
    end_match: int = 11,
    pipe=None,
    render: bool = True,
    visualize_xpos: bool = False
):
    """
    Visualizes the result of running Semgrex search on a string. The i-th element of
    the returned list is the HTML representation of the i-th sentence's dependency
    relationships. Only shows sentences that have a match on the Semgrex search.

    @param text: The string for which Semgrex search will be run on.
    @param semgrex_queries: A list of Semgrex queries to search for in the document.
    @param lang_code: A two letter language abbreviation for the language that the Stanza document is written in.
    @param start_match: Beginning of the splice for which to display elements with.
    @param end_match: End of the splice for which to display elements with.
    @param pipe: An NLP pipeline through which the text will be processed.
    @param render: A toggled option to render the HTML strings within the returned list.
    @param visualize_xpos: A toggled option to use xpos tags for part-of-speech labeling, defaulting to upos tags

    @return: A list of HTML strings representing the dependency relations of the doc object.
    """
    if pipe is None:
        nlp = Pipeline(lang_code, processors="tokenize, pos, lemma, depparse")
    else:
        nlp = pipe
    doc = nlp(text)
    return visualize_search_doc(
        doc,
        semgrex_queries,
        lang_code,
        start_match=start_match,
        end_match=end_match,
        render=render,
        visualize_xpos=visualize_xpos
    )


def adjust_dep_arrows(raw_html: str) -> str:
    """
    Default spaCy dependency visualizations have misaligned arrows. Fix arrows by aligning arrow ends and bodies
    to the word that they are directed to.

    @param raw_html: Dependency relation visualization generated HTML from displaCy
    @return: Edited HTML string with fixed arrow placements
    """

    HTML_ARROW_BEGINNING = '<g class="displacy-arrow">'
    HTML_ARROW_ENDING = "</g>"
    HTML_ARROW_ENDING_LEN = 6  # there are 2 newline chars after the arrow ending
    arrows_start_idx = find_nth(
        haystack=raw_html, needle='<g class="displacy-arrow">', n=1
    )
    words_html, arrows_html = (
        raw_html[:arrows_start_idx],
        raw_html[arrows_start_idx:],
    )  # separate html for words and arrows
    final_html = (
        words_html  # continually concatenate to this after processing each arrow
    )
    arrow_number = 1  # which arrow we're currently editing (1-indexed)
    start_idx, end_of_class_idx = (
        find_nth(haystack=arrows_html, needle=HTML_ARROW_BEGINNING, n=arrow_number),
        find_nth(haystack=arrows_html, needle=HTML_ARROW_ENDING, n=arrow_number),
    )
    while start_idx != -1:  # edit every arrow
        arrow_section = arrows_html[
            start_idx : end_of_class_idx + HTML_ARROW_ENDING_LEN
        ]  # slice a single svg arrow object
        if (
            arrow_section[-1] == "<"
        ):  # this is the last arrow in the HTML, don't cut the splice early
            arrow_section = arrows_html[start_idx:]
        edited_arrow_section = edit_dep_arrow(arrow_section)

        final_html = (
            final_html + edited_arrow_section
        )  # continually update html with new arrow html until done

        # Prepare for next iteration
        arrow_number += 1
        start_idx = find_nth(arrows_html, '<g class="displacy-arrow">', arrow_number)
        end_of_class_idx = find_nth(arrows_html, "</g>", arrow_number)
    return final_html


def edit_dep_arrow(arrow_html: str) -> str:
    """
    The formatting of a single displacy arrow in svg is the following:
    <g class="displacy-arrow">
        <path class="displacy-arc" id="arrow-c628889ffbf343e3848193a08606f10a-0-0" stroke-width="2px" d="M70,352.0 C70,177.0 390.0,177.0 390.0,352.0" fill="none" stroke="currentColor"/>
        <text dy="1.25em" style="font-size: 0.8em; letter-spacing: 1px">
            <textPath xlink:href="#arrow-c628889ffbf343e3848193a08606f10a-0-0" class="displacy-label" startOffset="50%" side="left" fill="currentColor" text-anchor="middle">csubj</textPath>
        </text>
        <path class="displacy-arrowhead" d="M70,354.0 L62,342.0 78,342.0" fill="currentColor"/>
    </g>

    We edit the 'd = ...' parts of the <path class ...> section to fix the arrow direction and length to round to
    the nearest 50 units, centering on each word's center. This is because the words start at x=50 and have spacing
    of 100, so each word is at an x-value that is a multiple of 50.

    @param arrow_html: Original SVG for a single displaCy arrow.
    @return: Edited SVG for the displaCy arrow, adjusting its placement
    """

    WORD_SPACING = 50  # words start at x=50 and are separated by 100s so their x values are multiples of 50
    M_OFFSET = 4  # length of 'd="M' that we search for to extract the number from d="M70, for instance
    ARROW_PIXEL_SIZE = 4
    first_d_idx, second_d_idx = (
        find_nth(arrow_html, 'd="M', 1),
        find_nth(arrow_html, 'd="M', 2),
    )  # find where d="M starts
    first_d_cutoff, second_d_cutoff = (
        arrow_html.find(",", first_d_idx),
        arrow_html.find(",", second_d_idx),
    )  # isolate the number after 'M' e.g. 'M70'
    # gives svg x values of arrow body starting position and arrowhead position
    arrow_position, arrowhead_position = (
        float(arrow_html[first_d_idx + M_OFFSET : first_d_cutoff]),
        float(arrow_html[second_d_idx + M_OFFSET : second_d_cutoff]),
    )
    # gives starting index of where 'fill="none"' or 'fill="currentColor"' begin, reference points to end the d= section
    first_fill_start_idx, second_fill_start_idx = (
        find_nth(arrow_html, "fill", n=1),
        find_nth(arrow_html, "fill", n=3),
    )

    # isolate the d= ... section to edit
    first_d, second_d = (
        arrow_html[first_d_idx:first_fill_start_idx],
        arrow_html[second_d_idx:second_fill_start_idx],
    )
    first_d_split, second_d_split = first_d.split(","), second_d.split(",")

    if (
        arrow_position == arrowhead_position
    ):  # This arrow is incoming onto the word, center the arrow/head to word center
        corrected_arrow_pos = corrected_arrowhead_pos = round_base(
            arrow_position, base=WORD_SPACING
        )

        # edit first_d  -- arrow body
        second_term = first_d_split[1].split(" ")[0] + " " + str(corrected_arrow_pos)
        first_d = (
            'd="M'
            + str(corrected_arrow_pos)
            + ","
            + second_term
            + ","
            + ",".join(first_d_split[2:])
        )

        # edit second_d  -- arrowhead
        second_term = (
            second_d_split[1].split(" ")[0]
            + " L"
            + str(corrected_arrowhead_pos - ARROW_PIXEL_SIZE)
        )
        third_term = (
            second_d_split[2].split(" ")[0]
            + " "
            + str(corrected_arrowhead_pos + ARROW_PIXEL_SIZE)
        )
        second_d = (
            'd="M'
            + str(corrected_arrowhead_pos)
            + ","
            + second_term
            + ","
            + third_term
            + ","
            + ",".join(second_d_split[3:])
        )
    else:  # This arrow is outgoing to another word, center the arrow/head to that word's center
        corrected_arrowhead_pos = round_base(arrowhead_position, base=WORD_SPACING)

        # edit first_d -- arrow body
        third_term = first_d_split[2].split(" ")[0] + " " + str(corrected_arrowhead_pos)
        fourth_term = (
            first_d_split[3].split(" ")[0] + " " + str(corrected_arrowhead_pos)
        )
        terms = [
            first_d_split[0],
            first_d_split[1],
            third_term,
            fourth_term,
        ] + first_d_split[4:]
        first_d = ",".join(terms)

        # edit second_d -- arrow head
        first_term = f'd="M{corrected_arrowhead_pos}'
        second_term = (
            second_d_split[1].split(" ")[0]
            + " L"
            + str(corrected_arrowhead_pos - ARROW_PIXEL_SIZE)
        )
        third_term = (
            second_d_split[2].split(" ")[0]
            + " "
            + str(corrected_arrowhead_pos + ARROW_PIXEL_SIZE)
        )
        terms = [first_term, second_term, third_term] + second_d_split[3:]
        second_d = ",".join(terms)
    # rebuild and return html from its individual sections
    return (
        arrow_html[:first_d_idx]
        + first_d
        + " "
        + arrow_html[first_fill_start_idx:second_d_idx]
        + second_d
        + " "
        + arrow_html[second_fill_start_idx:]
    )


def edit_html_overflow(html_string: str) -> str:
    """
    Adds to overflow and display settings to the SVG header to visualize overflowing HTML renderings in the
    Semgrex streamlit app. Prevents Semgrex search tags from being cut off at the bottom of visualizations.

    The opening of each HTML string looks similar to this; we add to the end of the SVG header.

    <svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" xml:lang="en" id="fa9446a525de4862b233007f26dbbecb-0" class="displacy" width="850" height="242.0" direction="ltr" style="max-width: none; height: 242.0px; color: #000000; background: #ffffff; font-family: Arial; direction: ltr">
    <style> .bolded{font-weight: bold;} </style>
    <text class="displacy-token" fill="currentColor" text-anchor="middle" y="182.0">
        <tspan class="bolded" fill="#66CCEE" x="50">Banning</tspan>

       <tspan class="displacy-tag" dy="2em" fill="currentColor" x="50">VERB</tspan>
      <tspan class="displacy-word" dy="2em" fill="#66CCEE" x=50>Act.</tspan>
    </text>

    @param html_string: HTML of the result of running Semgrex search on a text
    @return: Edited HTML to visualize the dependencies even in the case of overflow.
    """

    BUFFER_LEN = 14  # length of 'direction: ltr"'
    editing_start_idx = find_nth(html_string, "direction: ltr", n=1)
    SVG_HEADER_ADDITION = "overflow: visible; display: block"
    return (
        html_string[:editing_start_idx]
        + "; "
        + SVG_HEADER_ADDITION
        + html_string[editing_start_idx + BUFFER_LEN :]
    )


def main():
    """
    IMPORTANT: For the code in this module to run, you must have corenlp and Java installed on your machine. Additionally,
    set an environment variable CLASSPATH equal to the path of your corenlp directory.

    Example: CLASSPATH=C:\\Users\\Alex\\PycharmProjects\\pythonProject\\stanford-corenlp-4.5.0\\stanford-corenlp-4.5.0\\*
    """
    nlp = Pipeline("en", processors="tokenize,pos,lemma,depparse")
    doc = nlp(
        "Banning opal removed artifact decks from the meta. Banning tennis resulted in players banning people."
    )
    queries = [
        "{pos:NN}=object <obl {}=action",
        "{cpos:NOUN}=thing <obj {cpos:VERB}=action",
    ]

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--doc", type=Any, default=doc, help="Stanza document to process."
    )
    parser.add_argument(
        "--queries",
        type=List[str],
        default=queries,
        help="Semgrex queries to search for",
    )
    parser.add_argument(
        "--lang_code",
        type=str,
        default="en",
        help="Two letter abbreviation the document's language e.g. 'en' for English",
    )
    parser.add_argument(
        "--CLASSPATH",
        type=str,
        default="C:\\stanford-corenlp-4.5.2\\stanford-corenlp-4.5.2\\*",
        help="Path to your coreNLP directory",
    )
    args = parser.parse_args()
    os.environ["CLASSPATH"] = args.CLASSPATH
    try:
        res = visualize_search_doc(doc, queries, "en")
        print(res[0])  # first sentence's HTML
    except TypeError:
        raise TypeError(
            """For the code in this module to run, you must have corenlp and Java installed on your machine. 
            Once installed, you can pass in the path to your corenlp directory as a command-line argument named 
            "CLASSPATH". Alternatively, set an environment variable CLASSPATH equal to the path of your corenlp 
            directory."""
        )
    return


if __name__ == "__main__":
    main()



import argparse

import json
import torch
from tqdm import tqdm

from stanza.models.coref.model import CorefModel


if __name__ == "__main__":
    argparser = argparse.ArgumentParser()
    argparser.add_argument("experiment")
    argparser.add_argument("input_file")
    argparser.add_argument("output_file")
    argparser.add_argument("--config-file", default="config.toml")
    argparser.add_argument("--batch-size", type=int,
                           help="Adjust to override the config value if you're"
                                " experiencing out-of-memory issues")
    argparser.add_argument("--weights",
                           help="Path to file with weights to load."
                                " If not supplied, in the latest"
                                " weights of the experiment will be loaded;"
                                " if there aren't any, an error is raised.")
    args = argparser.parse_args()

    model = CorefModel.load_model(path=args.weights,
                                  map_location="cpu",
                                  ignore={"bert_optimizer", "general_optimizer",
                                          "bert_scheduler", "general_scheduler"})
    if args.batch_size:
        model.config.a_scoring_batch_size = args.batch_size
    model.training = False

    try:
        with open(args.input_file, encoding="utf-8") as fin:
            input_data = json.load(fin)
    except json.decoder.JSONDecodeError:
        # read the old jsonlines format if necessary
        with open(args.input_file, encoding="utf-8") as fin:
            text = "[" + ",\n".join(fin) + "]"
        input_data = json.loads(text)
    docs = [model.build_doc(doc) for doc in input_data]

    with torch.no_grad():
        for doc in tqdm(docs, unit="docs"):
            result = model.run(doc)
            doc["span_clusters"] = result.span_clusters
            doc["word_clusters"] = result.word_clusters

            for key in ("word2subword", "subwords", "word_id", "head2span"):
                del doc[key]

    with open(args.output_file, mode="w") as fout:
        for doc in docs:
            json.dump(doc, fout)



"""
Tests the conversion code for the lang_uk NER dataset
"""

import unittest
from stanza.utils.datasets.ner.convert_bsf_to_beios import convert_bsf, parse_bsf, BsfInfo

import pytest
pytestmark = [pytest.mark.travis, pytest.mark.pipeline]

class TestBsf2Beios(unittest.TestCase):
    
    def test_empty_markup(self):
        res = convert_bsf('', '')
        self.assertEqual('', res)

    def test_1line_markup(self):
        data = 'тележурналіст Василь'
        bsf_markup = 'T1	PERS 14 20	Василь'
        expected = '''тележурналіст O
Василь S-PERS'''
        self.assertEqual(expected, convert_bsf(data, bsf_markup))

    def test_1line_follow_markup(self):
        data = 'тележурналіст Василь .'
        bsf_markup = 'T1	PERS 14 20	Василь'
        expected = '''тележурналіст O
Василь S-PERS
. O'''
        self.assertEqual(expected, convert_bsf(data, bsf_markup))

    def test_1line_2tok_markup(self):
        data = 'тележурналіст Василь Нагірний .'
        bsf_markup = 'T1	PERS 14 29	Василь Нагірний'
        expected = '''тележурналіст O
Василь B-PERS
Нагірний E-PERS
. O'''
        self.assertEqual(expected, convert_bsf(data, bsf_markup))

    def test_1line_Long_tok_markup(self):
        data = 'А в музеї Гуцульщини і Покуття можна '
        bsf_markup = 'T12	ORG 4 30	музеї Гуцульщини і Покуття'
        expected = '''А O
в O
музеї B-ORG
Гуцульщини I-ORG
і I-ORG
Покуття E-ORG
можна O'''
        self.assertEqual(expected, convert_bsf(data, bsf_markup))

    def test_2line_2tok_markup(self):
        data = '''тележурналіст Василь Нагірний .
В івано-франківському видавництві «Лілея НВ» вийшла друком'''
        bsf_markup = '''T1	PERS 14 29	Василь Нагірний
T2	ORG 67 75	Лілея НВ'''
        expected = '''тележурналіст O
Василь B-PERS
Нагірний E-PERS
. O


В O
івано-франківському O
видавництві O
« O
Лілея B-ORG
НВ E-ORG
» O
вийшла O
друком O'''
        self.assertEqual(expected, convert_bsf(data, bsf_markup))

    def test_real_markup(self):
        data = '''Через напіввоєнний стан в Україні та збільшення телефонних терористичних погроз українці купуватимуть sim-карти тільки за паспортами .
Про це повідомив начальник управління зв'язків зі ЗМІ адміністрації Держспецзв'язку Віталій Кукса .
Він зауважив , що днями відомство опублікує проект змін до правил надання телекомунікаційних послуг , де будуть прописані норми ідентифікації громадян .
Абонентів , які на сьогодні вже мають sim-карту , за словами Віталія Кукси , реєструватимуть , коли ті звертатимуться в службу підтримки свого оператора мобільного зв'язку .
Однак мобільні оператори побоюються , що таке нововведення помітно зменшить продаж стартових пакетів , адже спеціалізовані магазини є лише у містах .
Відтак купити сімку в невеликих населених пунктах буде неможливо .
Крім того , нова процедура ідентифікації абонентів вимагатиме від операторів мобільного зв'язку додаткових витрат .
- Близько 90 % українських абонентів - це абоненти передоплати .
Якщо мова буде йти навіть про поетапну їх ідентифікацію , зробити це буде складно , довго і дорого .
Мобільним операторам доведеться йти на чималі витрати , пов'язані з укладанням і зберіганням договорів , веденням баз даних , - розповіла « Економічній правді » начальник відділу зв'язків з громадськістю « МТС-Україна » Вікторія Рубан .
'''
        bsf_markup = '''T1	LOC 26 33	Україні
T2	ORG 203 218	Держспецзв'язку
T3	PERS 219 232	Віталій Кукса
T4	PERS 449 462	Віталія Кукси
T5	ORG 1201 1219	Економічній правді
T6	ORG 1267 1278	МТС-Україна
T7	PERS 1281 1295	Вікторія Рубан
'''
        expected = '''Через O
напіввоєнний O
стан O
в O
Україні S-LOC
та O
збільшення O
телефонних O
терористичних O
погроз O
українці O
купуватимуть O
sim-карти O
тільки O
за O
паспортами O
. O


Про O
це O
повідомив O
начальник O
управління O
зв'язків O
зі O
ЗМІ O
адміністрації O
Держспецзв'язку S-ORG
Віталій B-PERS
Кукса E-PERS
. O


Він O
зауважив O
, O
що O
днями O
відомство O
опублікує O
проект O
змін O
до O
правил O
надання O
телекомунікаційних O
послуг O
, O
де O
будуть O
прописані O
норми O
ідентифікації O
громадян O
. O


Абонентів O
, O
які O
на O
сьогодні O
вже O
мають O
sim-карту O
, O
за O
словами O
Віталія B-PERS
Кукси E-PERS
, O
реєструватимуть O
, O
коли O
ті O
звертатимуться O
в O
службу O
підтримки O
свого O
оператора O
мобільного O
зв'язку O
. O


Однак O
мобільні O
оператори O
побоюються O
, O
що O
таке O
нововведення O
помітно O
зменшить O
продаж O
стартових O
пакетів O
, O
адже O
спеціалізовані O
магазини O
є O
лише O
у O
містах O
. O


Відтак O
купити O
сімку O
в O
невеликих O
населених O
пунктах O
буде O
неможливо O
. O


Крім O
того O
, O
нова O
процедура O
ідентифікації O
абонентів O
вимагатиме O
від O
операторів O
мобільного O
зв'язку O
додаткових O
витрат O
. O


- O
Близько O
90 O
% O
українських O
абонентів O
- O
це O
абоненти O
передоплати O
. O


Якщо O
мова O
буде O
йти O
навіть O
про O
поетапну O
їх O
ідентифікацію O
, O
зробити O
це O
буде O
складно O
, O
довго O
і O
дорого O
. O


Мобільним O
операторам O
доведеться O
йти O
на O
чималі O
витрати O
, O
пов'язані O
з O
укладанням O
і O
зберіганням O
договорів O
, O
веденням O
баз O
даних O
, O
- O
розповіла O
« O
Економічній B-ORG
правді E-ORG
» O
начальник O
відділу O
зв'язків O
з O
громадськістю O
« O
МТС-Україна S-ORG
» O
Вікторія B-PERS
Рубан E-PERS
. O'''
        self.assertEqual(expected, convert_bsf(data, bsf_markup))


class TestBsf(unittest.TestCase):

    def test_empty_bsf(self):
        self.assertEqual(parse_bsf(''), [])

    def test_empty2_bsf(self):
        self.assertEqual(parse_bsf(' \n \n'), [])

    def test_1line_bsf(self):
        bsf = 'T1	PERS 103 118	Василь Нагірний'
        res = parse_bsf(bsf)
        expected = BsfInfo('T1', 'PERS', 103, 118, 'Василь Нагірний')
        self.assertEqual(len(res), 1)
        self.assertEqual(res, [expected])

    def test_2line_bsf(self):
        bsf = '''T9	PERS 778 783	Карла
T10	MISC 814 819	міста'''
        res = parse_bsf(bsf)
        expected = [BsfInfo('T9', 'PERS', 778, 783, 'Карла'),
                    BsfInfo('T10', 'MISC', 814, 819, 'міста')]
        self.assertEqual(len(res), 2)
        self.assertEqual(res, expected)

    def test_multiline_bsf(self):
        bsf = '''T3	PERS 220 235	Андрієм Кіщуком
T4	MISC 251 285	А .
Kubler .
Світло і тіні маестро
T5	PERS 363 369	Кіблер'''
        res = parse_bsf(bsf)
        expected = [BsfInfo('T3', 'PERS', 220, 235, 'Андрієм Кіщуком'),
                    BsfInfo('T4', 'MISC', 251, 285, '''А .
Kubler .
Світло і тіні маестро'''),
                    BsfInfo('T5', 'PERS', 363, 369, 'Кіблер')]
        self.assertEqual(len(res), len(expected))
        self.assertEqual(res, expected)


if __name__ == '__main__':
    unittest.main()



import pytest

from stanza.tests import *

from stanza.models.common.vocab import EMPTY
from stanza.models.ner import utils

pytestmark = [pytest.mark.travis, pytest.mark.pipeline]

WORDS       = [["Unban",   "Mox",   "Opal"], ["Ragavan",  "is",     "red"], ["Urza",   "Lord",  "High", "Artificer", "goes", "infinite", "with",  "Thopter",    "Sword"]]
BIO_TAGS    = [["O",     "B-ART",  "I-ART"], ["B-MONKEY", "O",  "B-COLOR"], ["B-PER", "I-PER", "I-PER", "I-PER",        "O",        "O",    "O", "B-WEAPON", "B-WEAPON"]]
BIO_U_TAGS  = [["O",     "B_ART",  "I_ART"], ["B_MONKEY", "O",  "B_COLOR"], ["B_PER", "I_PER", "I_PER", "I_PER",        "O",        "O",    "O", "B_WEAPON", "B_WEAPON"]]
BIOES_TAGS  = [["O",     "B-ART",  "E-ART"], ["S-MONKEY", "O",  "S-COLOR"], ["B-PER", "I-PER", "I-PER", "E-PER",        "O",        "O",    "O", "S-WEAPON", "S-WEAPON"]]
# note the problem with not using BIO tags - the consecutive tags for thopter/sword get treated as one item
BASIC_TAGS  = [["O",       "ART",    "ART"], ["MONKEY",   "O",    "COLOR"], [  "PER",   "PER",   "PER",   "PER",        "O",        "O",    "O",   "WEAPON",   "WEAPON"]]
BASIC_BIOES = [["O",     "B-ART",  "E-ART"], ["S-MONKEY", "O",  "S-COLOR"], ["B-PER", "I-PER", "I-PER", "E-PER",        "O",        "O",    "O", "B-WEAPON", "E-WEAPON"]]
ALT_BIO     = [["O",    "B-MANA", "I-MANA"], ["B-CRE",    "O",        "O"], ["B-CRE", "I-CRE", "I-CRE", "I-CRE",        "O",        "O",    "O",    "B-ART",    "B-ART"]]
ALT_BIOES   = [["O",    "B-MANA", "E-MANA"], ["S-CRE",    "O",        "O"], ["B-CRE", "I-CRE", "I-CRE", "E-CRE",        "O",        "O",    "O",    "S-ART",    "S-ART"]]
NONE_BIO    = [["O",    "B-MANA", "I-MANA"], [None,      None,       None], ["B-CRE", "I-CRE", "I-CRE", "I-CRE",        "O",        "O",    "O",    "B-ART",    "B-ART"]]
NONE_BIOES  = [["O",    "B-MANA", "E-MANA"], [None,      None,       None], ["B-CRE", "I-CRE", "I-CRE", "E-CRE",        "O",        "O",    "O",    "S-ART",    "S-ART"]]
EMPTY_BIO   = [["O",    "B-MANA", "I-MANA"], [EMPTY,     EMPTY,     EMPTY], ["B-CRE", "I-CRE", "I-CRE", "I-CRE",        "O",        "O",    "O",    "B-ART",    "B-ART"]]

def test_normalize_empty_tags():
    sentences = [[(word[0], (word[1],)) for word in zip(*sentence)] for sentence in zip(WORDS, NONE_BIO)]
    new_sentences = utils.normalize_empty_tags(sentences)
    expected = [[(word[0], (word[1],)) for word in zip(*sentence)] for sentence in zip(WORDS, EMPTY_BIO)]
    assert new_sentences == expected

def check_reprocessed_tags(words, input_tags, expected_tags):
    sentences = [list(zip(x, y)) for x, y in zip(words, input_tags)]
    retagged = utils.process_tags(sentences=sentences, scheme="bioes")
    # process_tags selectively returns tuples or strings based on the input
    # so we don't need to fiddle with the expected output format here
    expected_retagged = [list(zip(x, y)) for x, y in zip(words, expected_tags)]
    assert retagged == expected_retagged

def test_process_tags_bio():
    check_reprocessed_tags(WORDS, BIO_TAGS, BIOES_TAGS)
    # check that the alternate version is correct as well
    # that way we can independently check the two layer version
    check_reprocessed_tags(WORDS, ALT_BIO, ALT_BIOES)

def test_process_tags_with_none():
    # if there is a block of tags with None in them, the Nones should be skipped over
    check_reprocessed_tags(WORDS, NONE_BIO, NONE_BIOES)

def merge_tags(*tags):
    merged_tags = [[tuple(x) for x in zip(*sentences)]   # combine tags such as ("O", "O"), ("B-ART", "B-MANA"), ...
                   for sentences in zip(*tags)]          # ... for each set of sentences
    return merged_tags

def test_combined_tags_bio():
    bio_tags = merge_tags(BIO_TAGS, ALT_BIO)
    expected = merge_tags(BIOES_TAGS, ALT_BIOES)
    check_reprocessed_tags(WORDS, bio_tags, expected)

def test_combined_tags_mixed():
    bio_tags = merge_tags(BIO_TAGS, ALT_BIOES)
    expected = merge_tags(BIOES_TAGS, ALT_BIOES)
    check_reprocessed_tags(WORDS, bio_tags, expected)

def test_process_tags_basic():
    check_reprocessed_tags(WORDS, BASIC_TAGS, BASIC_BIOES)

def test_process_tags_bioes():
    """
    This one should not change, naturally
    """
    check_reprocessed_tags(WORDS, BIOES_TAGS, BIOES_TAGS)
    check_reprocessed_tags(WORDS, BASIC_BIOES, BASIC_BIOES)

def run_flattened(fn, tags):
    return fn([x for x in y for y in tags])

def test_check_bio():
    assert     utils.is_bio_scheme([x for y in BIO_TAGS for x in y])
    assert not utils.is_bio_scheme([x for y in BIOES_TAGS for x in y])
    assert not utils.is_bio_scheme([x for y in BASIC_TAGS for x in y])
    assert not utils.is_bio_scheme([x for y in BASIC_BIOES for x in y])

def test_check_basic():
    assert not utils.is_basic_scheme([x for y in BIO_TAGS for x in y])
    assert not utils.is_basic_scheme([x for y in BIOES_TAGS for x in y])
    assert     utils.is_basic_scheme([x for y in BASIC_TAGS for x in y])
    assert not utils.is_basic_scheme([x for y in BASIC_BIOES for x in y])

def test_underscores():
    """
    Check that the methods work if the inputs are underscores instead of dashes
    """
    assert not utils.is_basic_scheme([x for y in BIO_U_TAGS for x in y])
    check_reprocessed_tags(WORDS, BIO_U_TAGS, BIOES_TAGS)

def test_merge_tags():
    """
    Check a few versions of the tag sequence merging
    """
    seq1     = [     "O",     "O",     "O", "B-FOO", "E-FOO",     "O"]
    seq2     = [ "S-FOO",     "O", "B-FOO", "E-FOO",     "O",     "O"]
    seq3     = [ "B-FOO", "E-FOO", "B-FOO", "E-FOO",     "O",     "O"]
    seq_err  = [     "O", "B-FOO",     "O", "B-FOO", "E-FOO",     "O"]
    seq_err2 = [     "O", "B-FOO",     "O", "B-FOO", "B-FOO",     "O"]
    seq_err3 = [     "O", "B-FOO",     "O", "B-FOO", "I-FOO",     "O"]
    seq_err4 = [     "O", "B-FOO",     "O", "B-FOO", "I-FOO", "I-FOO"]

    result = utils.merge_tags(seq1, seq2)
    expected = [ "S-FOO",     "O",     "O", "B-FOO", "E-FOO",     "O"]
    assert result == expected

    result = utils.merge_tags(seq2, seq1)
    expected = [ "S-FOO",     "O", "B-FOO", "E-FOO",     "O",     "O"]
    assert result == expected

    result = utils.merge_tags(seq1, seq3)
    expected = [ "B-FOO", "E-FOO",     "O", "B-FOO", "E-FOO",     "O"]
    assert result == expected

    with pytest.raises(ValueError):
        result = utils.merge_tags(seq1, seq_err)

    with pytest.raises(ValueError):
        result = utils.merge_tags(seq1, seq_err2)

    with pytest.raises(ValueError):
        result = utils.merge_tags(seq1, seq_err3)

    with pytest.raises(ValueError):
        result = utils.merge_tags(seq1, seq_err4)




"""
A short script to use a Stanza tokenizer to extract tokenized sentences from Wikipedia

The first step is to convert a Wikipedia dataset using Prof. Attardi's wikiextractor:
https://github.com/attardi/wikiextractor

This script then writes out sentences, one per line, whitespace separated
Some common issues with the tokenizer are accounted for by discarding those lines.

Also, to account for languages such as VI where whitespace occurs within words,
spaces are replaced with _  This should not cause any confusion, as any line with
a natural _ in has already been discarded.

for i in `echo A B C D E F G H I J K`; do nlprun "python3 stanza/utils/datasets/constituency/tokenize_wiki.py --output_file /u/nlp/data/constituency-parser/italian/2024_wiki_tokenization/it_wiki_tokenized_B$i.txt --lang it --max_len 120 --input_dir /u/nlp/data/Wikipedia/itwiki/B$i --tokenizer_model saved_models/tokenize/it_combined_tokenizer.pt --download_method None" -o /u/nlp/data/constituency-parser/italian/2024_wiki_tokenization/it_wiki_tokenized_B$i.out; done
"""

import argparse
import logging

import stanza
from stanza.models.common.bert_embedding import load_tokenizer, filter_data
from stanza.utils.datasets.constituency import selftrain_wiki
from stanza.utils.datasets.constituency.selftrain import add_length_args, tokenize_docs
from stanza.utils.get_tqdm import get_tqdm

tqdm = get_tqdm()

def parse_args():
    parser = argparse.ArgumentParser(
        description="Script that converts part of a wikipedia dump to silver standard trees"
    )
    parser.add_argument(
        '--output_file',
        default='vi_wiki_tokenized.txt',
        help='Where to write the tokenized lines'
    )
    parser.add_argument(
        '--lang',
        default='vi',
        help='Which language tools to use for tokenization and POS'
    )

    input_group = parser.add_mutually_exclusive_group(required=True)
    input_group.add_argument(
        '--input_dir',
        default=None,
        help='Path to the wikipedia dump after processing by wikiextractor'
    )
    input_group.add_argument(
        '--input_file',
        default=None,
        help='Path to a single file of the wikipedia dump after processing by wikiextractor'
    )
    parser.add_argument(
        '--bert_tokenizer',
        default=None,
        help='Which bert tokenizer (if any) to use to filter long sentences'
    )
    parser.add_argument(
        '--tokenizer_model',
        default=None,
        help='Use this model instead of the current Stanza tokenizer for this language'
    )
    parser.add_argument(
        '--download_method',
        default=None,
        help='Download pipeline models using this method (defaults to downloading updates from HF)'
    )
    add_length_args(parser)
    args = parser.parse_args()
    return args

def main():
    args = parse_args()
    if args.input_dir is not None:
        files = selftrain_wiki.list_wikipedia_files(args.input_dir)
    elif args.input_file is not None:
        files = [args.input_file]
    else:
        raise ValueError("Need to specify at least one file or directory!")

    if args.bert_tokenizer:
        tokenizer = load_tokenizer(args.bert_tokenizer)
        print("Max model length: %d" % tokenizer.model_max_length)
    pipeline_args = {}
    if args.tokenizer_model:
        pipeline_args["tokenize_model_path"] = args.tokenizer_model
    if args.download_method:
        pipeline_args["download_method"] = args.download_method
    pipe = stanza.Pipeline(args.lang, processors="tokenize", **pipeline_args)

    with open(args.output_file, "w", encoding="utf-8") as fout:
        for filename in tqdm(files):
            docs = selftrain_wiki.read_wiki_file(filename)
            text = tokenize_docs(docs, pipe, args.min_len, args.max_len)
            if args.bert_tokenizer:
                filtered = filter_data(args.bert_tokenizer, [x.split() for x in text], tokenizer, logging.DEBUG)
                text = [" ".join(x) for x in filtered]
            for line in text:
                fout.write(line)
                fout.write("\n")

if __name__ == '__main__':
    main()



import random
import logging
import torch

from stanza.models.common.bert_embedding import filter_data, needs_length_filter
from stanza.models.common.data import map_to_ids, get_long_tensor, sort_all
from stanza.models.common.vocab import PAD_ID, VOCAB_PREFIX
from stanza.models.pos.vocab import CharVocab, CompositeVocab, WordVocab
from stanza.models.ner.vocab import MultiVocab
from stanza.models.common.doc import *
from stanza.models.ner.utils import process_tags, normalize_empty_tags

logger = logging.getLogger('stanza')

class DataLoader:
    def __init__(self, doc, batch_size, args, pretrain=None, vocab=None, evaluation=False, preprocess_tags=True, bert_tokenizer=None, scheme=None, max_batch_words=None):
        self.max_batch_words = max_batch_words
        self.batch_size = batch_size
        self.args = args
        self.eval = evaluation
        self.shuffled = not self.eval
        self.doc = doc
        self.preprocess_tags = preprocess_tags

        data = self._load_doc(self.doc, scheme)

        # filter out the long sentences if bert is used
        if self.args.get('bert_model', None) and needs_length_filter(self.args['bert_model']):
            data = filter_data(self.args['bert_model'], data, bert_tokenizer)

        self.tags = [[w[1] for w in sent] for sent in data]
        # handle vocab
        self.pretrain = pretrain
        if vocab is None:
            self.vocab = self.init_vocab(data)
        else:
            self.vocab = vocab

        # filter and sample data
        if args.get('sample_train', 1.0) < 1.0 and not self.eval:
            keep = int(args['sample_train'] * len(data))
            data = random.sample(data, keep)
            logger.debug("Subsample training set with rate {:g}".format(args['sample_train']))

        data = self.preprocess(data, self.vocab, args)
        # shuffle for training
        if self.shuffled:
            random.shuffle(data)
        self.num_examples = len(data)

        # chunk into batches
        self.data = self.chunk_batches(data)
        logger.debug("{} batches created.".format(len(self.data)))

    def init_vocab(self, data):
        def from_model(model_filename):
            """ Try loading vocab from charLM model file. """
            state_dict = torch.load(model_filename, lambda storage, loc: storage, weights_only=True)
            if 'vocab' in state_dict:
                return state_dict['vocab']
            if 'model' in state_dict and 'vocab' in state_dict['model']:
                return state_dict['model']['vocab']
            raise ValueError("Cannot find vocab in charLM model file %s" % model_filename)

        if self.eval:
            raise AssertionError("Vocab must exist for evaluation.")
        if self.args['charlm']:
            charvocab = CharVocab.load_state_dict(from_model(self.args['charlm_forward_file']))
        else:
            charvocab = CharVocab(data, self.args['shorthand'])
        wordvocab = self.pretrain.vocab if self.pretrain is not None else None
        tag_data = [[(x[1],) for x in sentence] for sentence in data]
        tagvocab = CompositeVocab(tag_data, self.args['shorthand'], idx=0, sep=None)
        ignore = None
        if self.args['emb_finetune_known_only']:
            if self.pretrain is None:
                raise ValueError("Cannot train emb_finetune_known_only with no pretrain of known words")
            if self.args['lowercase']:
                ignore = set([w[0].lower() for sent in data for w in sent if w[0] not in wordvocab and w[0].lower() not in wordvocab])
            else:
                ignore = set([w[0] for sent in data for w in sent if w[0] not in wordvocab])
            logger.debug("Ignoring %d in the delta vocab as they did not appear in the original embedding", len(ignore))
        deltavocab = WordVocab(data, self.args['shorthand'], cutoff=1, lower=self.args['lowercase'], ignore=ignore)
        logger.debug("Creating delta vocab of size %s", len(deltavocab))
        vocabs = {'char': charvocab,
                  'delta': deltavocab,
                  'tag': tagvocab}
        if wordvocab is not None:
            vocabs['word'] = wordvocab
        vocab = MultiVocab(vocabs)
        return vocab

    def preprocess(self, data, vocab, args):
        processed = []
        if args.get('char_lowercase', False): # handle character case
            char_case = lambda x: x.lower()
        else:
            char_case = lambda x: x
        for sent_idx, sent in enumerate(data):
            processed_sent = [[w[0] for w in sent]]
            processed_sent += [[vocab['char'].map([char_case(x) for x in w[0]]) for w in sent]]
            processed_sent += [vocab['tag'].map([w[1] for w in sent])]
            processed.append(processed_sent)
        return processed

    def __len__(self):
        return len(self.data)

    def __getitem__(self, key):
        """ Get a batch with index. """
        if not isinstance(key, int):
            raise TypeError
        if key < 0 or key >= len(self.data):
            raise IndexError
        batch = self.data[key]
        batch_size = len(batch)
        batch = list(zip(*batch))
        assert len(batch) == 3 # words: List[List[int]], chars: List[List[List[int]]], tags: List[List[List[int]]]

        # sort sentences by lens for easy RNN operations
        sentlens = [len(x) for x in batch[0]]
        batch, orig_idx = sort_all(batch, sentlens)
        sentlens = [len(x) for x in batch[0]]

        # sort chars by lens for easy char-LM operations
        chars_forward, chars_backward, charoffsets_forward, charoffsets_backward, charlens = self.process_chars(batch[1])
        chars_sorted, char_orig_idx = sort_all([chars_forward, chars_backward, charoffsets_forward, charoffsets_backward], charlens)
        chars_forward, chars_backward, charoffsets_forward, charoffsets_backward = chars_sorted
        charlens = [len(sent) for sent in chars_forward]

        # sort words by lens for easy char-RNN operations
        batch_words = [w for sent in batch[1] for w in sent]
        wordlens = [len(x) for x in batch_words]
        batch_words, word_orig_idx = sort_all([batch_words], wordlens)
        batch_words = batch_words[0]
        wordlens = [len(x) for x in batch_words]

        words = batch[0]
        
        wordchars = get_long_tensor(batch_words, len(wordlens))
        wordchars_mask = torch.eq(wordchars, PAD_ID)
        chars_forward = get_long_tensor(chars_forward, batch_size, pad_id=self.vocab['char'].unit2id(' '))
        chars_backward = get_long_tensor(chars_backward, batch_size, pad_id=self.vocab['char'].unit2id(' '))
        chars = torch.cat([chars_forward.unsqueeze(0), chars_backward.unsqueeze(0)]) # padded forward and backward char idx
        charoffsets = [charoffsets_forward, charoffsets_backward] # idx for forward and backward lm to get word representation
        tags = get_long_tensor(batch[2], batch_size)

        return words, wordchars, wordchars_mask, chars, tags, orig_idx, word_orig_idx, char_orig_idx, sentlens, wordlens, charlens, charoffsets

    def __iter__(self):
        for i in range(self.__len__()):
            yield self.__getitem__(i)

    def _load_doc(self, doc, scheme):
        # preferentially load the MULTI_NER in case we are training /
        # testing a model with multiple layers of tags
        data = doc.get([TEXT, NER, MULTI_NER], as_sentences=True, from_token=True)
        data = [[[token[0], token[2]] if token[2] else [token[0], (token[1],)] for token in sentence] for sentence in data]
        if self.preprocess_tags: # preprocess tags
            if scheme is None:
                data = process_tags(data, self.args.get('scheme', 'bio'))
            data = normalize_empty_tags(data)
        return data

    def process_chars(self, sents):
        start_id, end_id = self.vocab['char'].unit2id('\n'), self.vocab['char'].unit2id(' ') # special token
        start_offset, end_offset = 1, 1
        chars_forward, chars_backward, charoffsets_forward, charoffsets_backward = [], [], [], []
        # get char representation for each sentence
        for sent in sents:
            chars_forward_sent, chars_backward_sent, charoffsets_forward_sent, charoffsets_backward_sent = [start_id], [start_id], [], []
            # forward lm
            for word in sent:
                chars_forward_sent += word
                charoffsets_forward_sent = charoffsets_forward_sent + [len(chars_forward_sent)] # add each token offset in the last for forward lm
                chars_forward_sent += [end_id]
            # backward lm
            for word in sent[::-1]:
                chars_backward_sent += word[::-1]
                charoffsets_backward_sent = [len(chars_backward_sent)] + charoffsets_backward_sent # add each offset in the first for backward lm
                chars_backward_sent += [end_id]
            # store each sentence
            chars_forward.append(chars_forward_sent)
            chars_backward.append(chars_backward_sent)
            charoffsets_forward.append(charoffsets_forward_sent)
            charoffsets_backward.append(charoffsets_backward_sent)
        charlens = [len(sent) for sent in chars_forward] # forward lm and backward lm should have the same lengths
        return chars_forward, chars_backward, charoffsets_forward, charoffsets_backward, charlens

    def reshuffle(self):
        data = [y for x in self.data for y in x]
        random.shuffle(data)
        self.data = self.chunk_batches(data)

    def chunk_batches(self, data):
        if self.max_batch_words is None:
            return [data[i:i+self.batch_size] for i in range(0, len(data), self.batch_size)]
        batches = []
        next_batch = []
        for item in data:
            next_batch.append(item)
            if len(next_batch) >= self.batch_size:
                batches.append(next_batch)
                next_batch = []
            if sum(len(x[0]) for x in next_batch) >= self.max_batch_words:
                batches.append(next_batch)
                next_batch = []
        if len(next_batch) > 0:
            batches.append(next_batch)
        return batches



"""
This script allows for training or testing on dev / test of the UD lemmatizer.

If run with a single treebank name, it will train or test that treebank.
If run with ud_all or all_ud, it will iterate over all UD treebanks it can find.

Mode can be set to train&dev with --train, to dev set only
with --score_dev, and to test set only with --score_test.

Treebanks are specified as a list.  all_ud or ud_all means to look for
all UD treebanks.

Extra arguments are passed to the lemmatizer.  In case the run script
itself is shadowing arguments, you can specify --extra_args as a
parameter to mark where the lemmatizer arguments start.
"""

import logging
import os

from stanza.models import identity_lemmatizer
from stanza.models import lemmatizer

from stanza.utils.training import common
from stanza.utils.training.common import Mode, add_charlm_args, build_lemma_charlm_args, choose_lemma_charlm

from stanza.utils.datasets.prepare_lemma_treebank import check_lemmas
import stanza.utils.datasets.prepare_lemma_classifier as prepare_lemma_classifier

logger = logging.getLogger('stanza')

def add_lemma_args(parser):
    add_charlm_args(parser)

    parser.add_argument('--lemma_classifier', dest='lemma_classifier', action='store_true', default=None,
                        help="Don't use the lemma classifier datasets.  Default is to build lemma classifier as part of the original lemmatizer if the charlm is used")
    parser.add_argument('--no_lemma_classifier', dest='lemma_classifier', action='store_false',
                        help="Don't use the lemma classifier datasets.  Default is to build lemma classifier as part of the original lemmatizer if the charlm is used")

def build_model_filename(paths, short_name, command_args, extra_args):
    """
    Figure out what the model savename will be, taking into account the model settings.

    Useful for figuring out if the model already exists

    None will represent that there is no expected save_name
    """
    short_language, dataset = short_name.split("_", 1)

    lemma_dir      = paths["LEMMA_DATA_DIR"]
    train_file     = f"{lemma_dir}/{short_name}.train.in.conllu"

    if not os.path.exists(train_file):
        logger.debug("Treebank %s is not prepared for training the lemmatizer.  Could not find any training data at %s  Cannot figure out the expected save_name without looking at the data, but a later step in the process will skip the training anyway" % (short_name, train_file))
        return None

    has_lemmas = check_lemmas(train_file)
    if not has_lemmas:
        return None

    # TODO: can avoid downloading the charlm at this point, since we
    # might not even be training
    charlm_args = build_lemma_charlm_args(short_language, dataset, command_args.charlm)

    train_args = ["--train_file", train_file,
                  "--shorthand", short_name,
                  "--mode", "train"]
    train_args = train_args + charlm_args + extra_args
    args = lemmatizer.parse_args(train_args)
    save_name = lemmatizer.build_model_filename(args)
    return save_name

def run_treebank(mode, paths, treebank, short_name, command_args, extra_args):
    short_language, dataset = short_name.split("_", 1)

    lemma_dir      = paths["LEMMA_DATA_DIR"]
    train_file     = f"{lemma_dir}/{short_name}.train.in.conllu"
    dev_in_file    = f"{lemma_dir}/{short_name}.dev.in.conllu"
    dev_pred_file  = f"{lemma_dir}/{short_name}.dev.pred.conllu"
    test_in_file   = f"{lemma_dir}/{short_name}.test.in.conllu"
    test_pred_file = f"{lemma_dir}/{short_name}.test.pred.conllu"

    charlm_args = build_lemma_charlm_args(short_language, dataset, command_args.charlm)

    if not os.path.exists(train_file):
        logger.error("Treebank %s is not prepared for training the lemmatizer.  Could not find any training data at %s  Skipping..." % (treebank, train_file))
        return

    has_lemmas = check_lemmas(train_file)
    if not has_lemmas:
        logger.info("Treebank " + treebank + " (" + short_name +
                    ") has no lemmas.  Using identity lemmatizer")
        if mode == Mode.TRAIN or mode == Mode.SCORE_DEV:
            train_args = ["--train_file", train_file,
                          "--eval_file", dev_in_file,
                          "--gold_file", dev_in_file,
                          "--shorthand", short_name]
            if command_args.save_output:
                train_args.extend(["--output_file", dev_pred_file])
            logger.info("Running identity lemmatizer for {} with args {}".format(treebank, train_args))
            identity_lemmatizer.main(train_args)
        elif mode == Mode.SCORE_TEST:
            train_args = ["--train_file", train_file,
                          "--eval_file", test_in_file,
                          "--gold_file", test_in_file,
                          "--shorthand", short_name]
            if command_args.save_output:
                train_args.extend(["--output_file", test_pred_file])
            logger.info("Running identity lemmatizer for {} with args {}".format(treebank, train_args))
            identity_lemmatizer.main(train_args)            
    else:
        if mode == Mode.TRAIN:
            # ('UD_Czech-PDT', 'UD_Russian-SynTagRus', 'UD_German-HDT')
            if short_name in ('cs_pdt', 'ru_syntagrus', 'de_hdt'):
                num_epochs = "30"
            else:
                num_epochs = "60"

            train_args = ["--train_file", train_file,
                          "--eval_file", dev_in_file,
                          "--shorthand", short_name,
                          "--num_epoch", num_epochs,
                          "--mode", "train"]
            train_args = train_args + charlm_args + extra_args
            logger.info("Running train lemmatizer for {} with args {}".format(treebank, train_args))
            lemmatizer.main(train_args)

        if mode == Mode.SCORE_DEV or mode == Mode.TRAIN:
            dev_args = ["--eval_file", dev_in_file,
                        "--shorthand", short_name,
                        "--mode", "predict"]
            if command_args.save_output:
                train_args.extend(["--output_file", dev_pred_file])
            dev_args = dev_args + charlm_args + extra_args
            logger.info("Running dev lemmatizer for {} with args {}".format(treebank, dev_args))
            lemmatizer.main(dev_args)

        if mode == Mode.SCORE_TEST:
            test_args = ["--eval_file", test_in_file,
                         "--shorthand", short_name,
                         "--mode", "predict"]
            if command_args.save_output:
                train_args.extend(["--output_file", test_pred_file])
            test_args = test_args + charlm_args + extra_args
            logger.info("Running test lemmatizer for {} with args {}".format(treebank, test_args))
            lemmatizer.main(test_args)

        use_lemma_classifier = command_args.lemma_classifier
        if use_lemma_classifier is None:
            use_lemma_classifier = command_args.charlm is not None
        use_lemma_classifier = use_lemma_classifier and short_name in prepare_lemma_classifier.DATASET_MAPPING
        if use_lemma_classifier and mode == Mode.TRAIN:
            # some installations may not have transformers,
            # so we bury the lemma_classifier import in the codepath
            # which actually needs it
            from stanza.models.lemma import attach_lemma_classifier
            from stanza.utils.training import run_lemma_classifier

            lc_charlm_args = ['--no_charlm'] if command_args.charlm is None else ['--charlm', command_args.charlm]
            lemma_classifier_args = [treebank] + lc_charlm_args
            if command_args.force:
                lemma_classifier_args.append('--force')
            run_lemma_classifier.main(lemma_classifier_args)

            save_name = build_model_filename(paths, short_name, command_args, extra_args)
            # TODO: use a temp path for the lemma_classifier or keep it somewhere
            attach_args = ['--input', save_name,
                           '--output', save_name,
                           '--classifier', 'saved_models/lemma_classifier/%s_lemma_classifier.pt' % short_name]
            attach_lemma_classifier.main(attach_args)

            # now we rerun the dev set - the HI in particular demonstrates some good improvement
            lemmatizer.main(dev_args)

def main():
    common.main(run_treebank, "lemma", "lemmatizer", add_lemma_args, sub_argparse=lemmatizer.build_argparse(), build_model_filename=build_model_filename, choose_charlm_method=choose_lemma_charlm)

if __name__ == "__main__":
    main()




import pytest

import io
import os
import xml.etree.ElementTree as ET

from stanza.utils.datasets.ner.convert_nkjp import MORPH_FILE, NER_FILE, extract_entities_from_subfolder, extract_entities_from_sentence, extract_unassigned_subfolder_entities

pytestmark = [pytest.mark.pipeline, pytest.mark.travis]

EXPECTED_ENTITIES = {
    '1-p': {
        '1.39-s': [{'ent_id': 'named_1.39-s_n1', 'index': 0, 'orth': 'Sił Zbrojnych', 'ner_type': 'orgName', 'ner_subtype': None, 'targets': ['1.37-seg', '1.38-seg']}],
        '1.56-s': [],
        '1.79-s': []
    },
    '2-p': {
        '2.30-s': [],
        '2.45-s': []
    },
    '3-p': {
        '3.70-s': []
    }
}


@pytest.fixture(scope="module")
def dataset(tmp_path_factory):
    dataset_path = tmp_path_factory.mktemp("nkjp_dataset")
    sample_path = dataset_path / "sample"
    os.mkdir(sample_path)
    ann_path = sample_path / NER_FILE
    with open(ann_path, "w", encoding="utf-8") as fout:
        fout.write(SAMPLE_ANN)
    morph_path = sample_path / MORPH_FILE
    with open(morph_path, "w", encoding="utf-8") as fout:
        fout.write(SAMPLE_MORPHO)
    return dataset_path

EXPECTED_TOKENS = [
    {'seg_id': '1.1-seg', 'i': 0, 'orth': '2', 'text': '2', 'tag': '_', 'ner': 'O', 'ner_subtype': None},
    {'seg_id': '1.37-seg', 'i': 36, 'orth': 'Sił', 'text': 'Sił', 'tag': '_', 'ner': 'B-orgName', 'ner_subtype': None},
    {'seg_id': '1.38-seg', 'i': 37, 'orth': 'Zbrojnych', 'text': 'Zbrojnych', 'tag': '_', 'ner': 'I-orgName', 'ner_subtype': None},
]

def test_extract_entities_from_subfolder(dataset):
    entities = extract_entities_from_subfolder("sample", dataset)
    assert len(entities) == 1
    assert len(entities['1-p']) == 1
    assert len(entities['1-p']['1.39-s']) == 39
    assert entities['1-p']['1.39-s']['1.1-seg'] == EXPECTED_TOKENS[0]
    assert entities['1-p']['1.39-s']['1.37-seg'] == EXPECTED_TOKENS[1]
    assert entities['1-p']['1.39-s']['1.38-seg'] == EXPECTED_TOKENS[2]


def test_extract_unassigned(dataset):
    entities = extract_unassigned_subfolder_entities("sample", dataset)
    assert entities == EXPECTED_ENTITIES

SENTENCE_SAMPLE = """
          <s xmlns="http://www.tei-c.org/ns/1.0" xmlns:xi="http://www.w3.org/2001/XInclude" xml:id="named_1.39-s" corresp="ann_morphosyntax.xml#morph_1.39-s">
            <seg xml:id="named_1.39-s_n1">
              <fs type="named">
                <f name="type">
                  <symbol value="orgName"/>
                </f>
                <f name="orth">
                  <string>Si&#322; Zbrojnych</string>
                </f>
                <f name="base">
                  <string>Si&#322;y Zbrojne</string>
                </f>
                <f name="certainty">
                  <symbol value="high"/>
                </f>
              </fs>
              <ptr target="ann_morphosyntax.xml#morph_1.37-seg"/>
              <ptr target="ann_morphosyntax.xml#morph_1.38-seg"/>
            </seg>
          </s>
""".strip()


EMPTY_SENTENCE = """<s xml:id="named_1.56-s" corresp="ann_morphosyntax.xml#morph_1.56-s"/>"""

def test_extract_entities_from_sentence():
    rt = ET.fromstring(SENTENCE_SAMPLE)
    entities = extract_entities_from_sentence(rt)
    assert entities == EXPECTED_ENTITIES['1-p']['1.39-s']

    rt = ET.fromstring(EMPTY_SENTENCE)
    entities = extract_entities_from_sentence(rt)
    assert entities == []



# picked completely at random, one sample file for testing:
# 610-1-000248/ann_named.xml
# only the first sentence is used in the morpho file
SAMPLE_ANN = """
<?xml version='1.0' encoding='UTF-8'?>
<teiCorpus xmlns:xi="http://www.w3.org/2001/XInclude" xmlns="http://www.tei-c.org/ns/1.0">
  <xi:include href="NKJP_1M_header.xml"/>
  <TEI>
    <xi:include href="header.xml"/>
    <text xml:lang="pl">
      <body>
        <p xml:id="named_1-p" corresp="ann_morphosyntax.xml#morph_1-p">
          <s xml:id="named_1.39-s" corresp="ann_morphosyntax.xml#morph_1.39-s">
            <seg xml:id="named_1.39-s_n1">
              <fs type="named">
                <f name="type">
                  <symbol value="orgName"/>
                </f>
                <f name="orth">
                  <string>Sił Zbrojnych</string>
                </f>
                <f name="base">
                  <string>Siły Zbrojne</string>
                </f>
                <f name="certainty">
                  <symbol value="high"/>
                </f>
              </fs>
              <ptr target="ann_morphosyntax.xml#morph_1.37-seg"/>
              <ptr target="ann_morphosyntax.xml#morph_1.38-seg"/>
            </seg>
          </s>
          <s xml:id="named_1.56-s" corresp="ann_morphosyntax.xml#morph_1.56-s"/>
          <s xml:id="named_1.79-s" corresp="ann_morphosyntax.xml#morph_1.79-s"/>
        </p>
        <p xml:id="named_2-p" corresp="ann_morphosyntax.xml#morph_2-p">
          <s xml:id="named_2.30-s" corresp="ann_morphosyntax.xml#morph_2.30-s"/>
          <s xml:id="named_2.45-s" corresp="ann_morphosyntax.xml#morph_2.45-s"/>
        </p>
        <p xml:id="named_3-p" corresp="ann_morphosyntax.xml#morph_3-p">
          <s xml:id="named_3.70-s" corresp="ann_morphosyntax.xml#morph_3.70-s"/>
        </p>
      </body>
    </text>
  </TEI>
</teiCorpus>
""".lstrip()



SAMPLE_MORPHO = """
<?xml version="1.0" encoding="UTF-8"?>
<!-- w indeksach elementów wciąganych mogą zdarzyć się nieciągłości (z alternatyw segmentacyjnych)  --><teiCorpus xmlns="http://www.tei-c.org/ns/1.0" xmlns:nkjp="http://www.nkjp.pl/ns/1.0" xmlns:xi="http://www.w3.org/2001/XInclude">
 <xi:include href="NKJP_1M_header.xml"/>
 <TEI>
  <xi:include href="header.xml"/>
  <text>
   <body>
    <!-- morph_1-p is akapit 7626 with instances (akapit_transzy-s) 15244, 15269 in batches (transza-s) 1525, 1528 resp. -->
    <p corresp="ann_segmentation.xml#segm_1-p" xml:id="morph_1-p">
     <s corresp="ann_segmentation.xml#segm_1.39-s" xml:id="morph_1.39-s">
      <seg corresp="ann_segmentation.xml#segm_1.1-seg" xml:id="morph_1.1-seg">
       <fs type="morph">
        <f name="orth">
         <string>2</string>
        </f>
        <!-- 2 [0,1] -->
        <f name="interps">
         <fs type="lex" xml:id="morph_1.1.1-lex">
          <f name="base">
           <string/>
          </f>
          <f name="ctag">
           <symbol value="ign"/>
          </f>
          <f name="msd">
           <symbol value="" xml:id="morph_1.1.1.1-msd"/>
          </f>
         </fs>
         <fs type="lex" xml:id="morph_1.1.2-lex">
          <f name="base">
           <string>2</string>
          </f>
          <f name="ctag">
           <symbol value="adj"/>
          </f>
          <f name="msd">
           <symbol nkjp:manual="true" value="sg:nom:n:pos" xml:id="morph_1.1.2.1-msd"/>
          </f>
         </fs>
        </f>
        <f name="disamb">
         <fs feats="#an8003" type="tool_report">
          <f fVal="#morph_1.1.2.1-msd" name="choice"/>
          <f name="interpretation">
           <string>2:adj:sg:nom:n:pos</string>
          </f>
         </fs>
        </f>
       </fs>
      </seg>
      <seg corresp="ann_segmentation.xml#segm_1.2-seg" xml:id="morph_1.2-seg">
       <fs type="morph">
        <f name="orth">
         <string>.</string>
        </f>
        <!-- . [1,1] -->
        <f name="nps">
         <binary value="true"/>
        </f>
        <f name="interps">
         <fs type="lex" xml:id="morph_1.2.1-lex">
          <f name="base">
           <string>.</string>
          </f>
          <f name="ctag">
           <symbol value="interp"/>
          </f>
          <f name="msd">
           <symbol value="" xml:id="morph_1.2.1.1-msd"/>
          </f>
         </fs>
        </f>
        <f name="disamb">
         <fs feats="#an8003" type="tool_report">
          <f fVal="#morph_1.2.1.1-msd" name="choice"/>
          <f name="interpretation">
           <string>.:interp</string>
          </f>
         </fs>
        </f>
       </fs>
      </seg>
      <seg corresp="ann_segmentation.xml#segm_1.3-seg" xml:id="morph_1.3-seg">
       <fs type="morph">
        <f name="orth">
         <string>Wezwanie</string>
        </f>
        <!-- Wezwanie [3,8] -->
        <f name="interps">
         <fs type="lex" xml:id="morph_1.3.1-lex">
          <f name="base">
           <string>wezwanie</string>
          </f>
          <f name="ctag">
           <symbol value="subst"/>
          </f>
          <f name="msd">
           <vAlt>
            <symbol value="sg:nom:n" xml:id="morph_1.3.1.1-msd"/>
            <symbol value="sg:acc:n" xml:id="morph_1.3.1.2-msd"/>
            <symbol value="sg:voc:n" xml:id="morph_1.3.1.3-msd"/>
           </vAlt>
          </f>
         </fs>
         <fs type="lex" xml:id="morph_1.3.2-lex">
          <f name="base">
           <string>wezwać</string>
          </f>
          <f name="ctag">
           <symbol value="ger"/>
          </f>
          <f name="msd">
           <vAlt>
            <symbol value="sg:nom:n:perf:aff" xml:id="morph_1.3.2.1-msd"/>
            <symbol value="sg:acc:n:perf:aff" xml:id="morph_1.3.2.2-msd"/>
           </vAlt>
          </f>
         </fs>
        </f>
        <f name="disamb">
         <fs feats="#an8003" type="tool_report">
          <f fVal="#morph_1.3.1.2-msd" name="choice"/>
          <f name="interpretation">
           <string>wezwanie:subst:sg:acc:n</string>
          </f>
         </fs>
        </f>
       </fs>
      </seg>
      <seg corresp="ann_segmentation.xml#segm_1.4-seg" xml:id="morph_1.4-seg">
       <fs type="morph">
        <f name="orth">
         <string>,</string>
        </f>
        <!-- , [11,1] -->
        <f name="nps">
         <binary value="true"/>
        </f>
        <f name="interps">
         <fs type="lex" xml:id="morph_1.4.1-lex">
          <f name="base">
           <string>,</string>
          </f>
          <f name="ctag">
           <symbol value="interp"/>
          </f>
          <f name="msd">
           <symbol value="" xml:id="morph_1.4.1.1-msd"/>
          </f>
         </fs>
        </f>
        <f name="disamb">
         <fs feats="#an8003" type="tool_report">
          <f fVal="#morph_1.4.1.1-msd" name="choice"/>
          <f name="interpretation">
           <string>,:interp</string>
          </f>
         </fs>
        </f>
       </fs>
      </seg>
      <seg corresp="ann_segmentation.xml#segm_1.5-seg" xml:id="morph_1.5-seg">
       <fs type="morph">
        <f name="orth">
         <string>o</string>
        </f>
        <!-- o [13,1] -->
        <f name="interps">
         <fs type="lex" xml:id="morph_1.5.1-lex">
          <f name="base">
           <string>o</string>
          </f>
          <f name="ctag">
           <symbol value="interj"/>
          </f>
          <f name="msd">
           <symbol value="" xml:id="morph_1.5.1.1-msd"/>
          </f>
         </fs>
         <fs type="lex" xml:id="morph_1.5.2-lex">
          <f name="base">
           <string>o</string>
          </f>
          <f name="ctag">
           <symbol value="prep"/>
          </f>
          <f name="msd">
           <vAlt>
            <symbol value="acc" xml:id="morph_1.5.2.1-msd"/>
            <symbol value="loc" xml:id="morph_1.5.2.2-msd"/>
           </vAlt>
          </f>
         </fs>
         <fs type="lex" xml:id="morph_1.5.3-lex">
          <f name="base">
           <string>ojciec</string>
          </f>
          <f name="ctag">
           <symbol value="brev"/>
          </f>
          <f name="msd">
           <symbol value="pun" xml:id="morph_1.5.3.1-msd"/>
          </f>
         </fs>
        </f>
        <f name="disamb">
         <fs feats="#an8003" type="tool_report">
          <f fVal="#morph_1.5.2.2-msd" name="choice"/>
          <f name="interpretation">
           <string>o:prep:loc</string>
          </f>
         </fs>
        </f>
       </fs>
      </seg>
      <seg corresp="ann_segmentation.xml#segm_1.6-seg" xml:id="morph_1.6-seg">
       <fs type="morph">
        <f name="orth">
         <string>którym</string>
        </f>
        <!-- którym [15,6] -->
        <f name="interps">
         <fs type="lex" xml:id="morph_1.6.1-lex">
          <f name="base">
           <string>który</string>
          </f>
          <f name="ctag">
           <symbol value="adj"/>
          </f>
          <f name="msd">
           <vAlt>
            <symbol value="sg:inst:m1:pos" xml:id="morph_1.6.1.1-msd"/>
            <symbol value="sg:inst:m2:pos" xml:id="morph_1.6.1.2-msd"/>
            <symbol value="sg:inst:m3:pos" xml:id="morph_1.6.1.3-msd"/>
            <symbol value="sg:inst:n:pos" xml:id="morph_1.6.1.4-msd"/>
            <symbol value="sg:loc:m1:pos" xml:id="morph_1.6.1.5-msd"/>
            <symbol value="sg:loc:m2:pos" xml:id="morph_1.6.1.6-msd"/>
            <symbol value="sg:loc:m3:pos" xml:id="morph_1.6.1.7-msd"/>
            <symbol value="sg:loc:n:pos" xml:id="morph_1.6.1.8-msd"/>
            <symbol value="pl:dat:m1:pos" xml:id="morph_1.6.1.9-msd"/>
            <symbol value="pl:dat:m2:pos" xml:id="morph_1.6.1.10-msd"/>
            <symbol value="pl:dat:m3:pos" xml:id="morph_1.6.1.11-msd"/>
            <symbol value="pl:dat:f:pos" xml:id="morph_1.6.1.12-msd"/>
            <symbol value="pl:dat:n:pos" xml:id="morph_1.6.1.13-msd"/>
           </vAlt>
          </f>
         </fs>
        </f>
        <f name="disamb">
         <fs feats="#an8003" type="tool_report">
          <f fVal="#morph_1.6.1.8-msd" name="choice"/>
          <f name="interpretation">
           <string>który:adj:sg:loc:n:pos</string>
          </f>
         </fs>
        </f>
       </fs>
      </seg>
      <seg corresp="ann_segmentation.xml#segm_1.7-seg" xml:id="morph_1.7-seg">
       <fs type="morph">
        <f name="orth">
         <string>mowa</string>
        </f>
        <!-- mowa [22,4] -->
        <f name="interps">
         <fs type="lex" xml:id="morph_1.7.1-lex">
          <f name="base">
           <string>mowa</string>
          </f>
          <f name="ctag">
           <symbol value="subst"/>
          </f>
          <f name="msd">
           <symbol value="sg:nom:f" xml:id="morph_1.7.1.1-msd"/>
          </f>
         </fs>
        </f>
        <f name="disamb">
         <fs feats="#an8003" type="tool_report">
          <f fVal="#morph_1.7.1.1-msd" name="choice"/>
          <f name="interpretation">
           <string>mowa:subst:sg:nom:f</string>
          </f>
         </fs>
        </f>
       </fs>
      </seg>
      <seg corresp="ann_segmentation.xml#segm_1.8-seg" xml:id="morph_1.8-seg">
       <fs type="morph">
        <f name="orth">
         <string>w</string>
        </f>
        <!-- w [27,1] -->
        <f name="interps">
         <fs type="lex" xml:id="morph_1.8.1-lex">
          <f name="base">
           <string>w</string>
          </f>
          <f name="ctag">
           <symbol value="prep"/>
          </f>
          <f name="msd">
           <vAlt>
            <symbol value="acc:nwok" xml:id="morph_1.8.1.1-msd"/>
            <symbol value="loc:nwok" xml:id="morph_1.8.1.2-msd"/>
           </vAlt>
          </f>
         </fs>
         <fs type="lex" xml:id="morph_1.8.2-lex">
          <f name="base">
           <string>wiek</string>
          </f>
          <f name="ctag">
           <symbol value="brev"/>
          </f>
          <f name="msd">
           <symbol value="pun" xml:id="morph_1.8.2.1-msd"/>
          </f>
         </fs>
         <fs type="lex" xml:id="morph_1.8.3-lex">
          <f name="base">
           <string>wielki</string>
          </f>
          <f name="ctag">
           <symbol value="brev"/>
          </f>
          <f name="msd">
           <symbol value="pun" xml:id="morph_1.8.3.1-msd"/>
          </f>
         </fs>
         <fs type="lex" xml:id="morph_1.8.4-lex">
          <f name="base">
           <string>wiersz</string>
          </f>
          <f name="ctag">
           <symbol value="brev"/>
          </f>
          <f name="msd">
           <symbol value="pun" xml:id="morph_1.8.4.1-msd"/>
          </f>
         </fs>
         <fs type="lex" xml:id="morph_1.8.5-lex">
          <f name="base">
           <string>wieś</string>
          </f>
          <f name="ctag">
           <symbol value="brev"/>
          </f>
          <f name="msd">
           <symbol value="pun" xml:id="morph_1.8.5.1-msd"/>
          </f>
         </fs>
         <fs type="lex" xml:id="morph_1.8.6-lex">
          <f name="base">
           <string>wyspa</string>
          </f>
          <f name="ctag">
           <symbol value="brev"/>
          </f>
          <f name="msd">
           <symbol value="pun" xml:id="morph_1.8.6.1-msd"/>
          </f>
         </fs>
        </f>
        <f name="disamb">
         <fs feats="#an8003" type="tool_report">
          <f fVal="#morph_1.8.1.2-msd" name="choice"/>
          <f name="interpretation">
           <string>w:prep:loc:nwok</string>
          </f>
         </fs>
        </f>
       </fs>
      </seg>
      <seg corresp="ann_segmentation.xml#segm_1.9-seg" xml:id="morph_1.9-seg">
       <fs type="morph">
        <f name="orth">
         <string>ust</string>
        </f>
        <!-- ust [29,3] -->
        <f name="interps">
         <fs type="lex" xml:id="morph_1.9.1-lex">
          <f name="base">
           <string>usta</string>
          </f>
          <f name="ctag">
           <symbol value="subst"/>
          </f>
          <f name="msd">
           <symbol value="pl:gen:n" xml:id="morph_1.9.1.1-msd"/>
          </f>
         </fs>
         <fs type="lex" xml:id="morph_1.9.2-lex">
          <f name="base">
           <string>ustęp</string>
          </f>
          <f name="ctag">
           <symbol value="brev"/>
          </f>
          <f name="msd">
           <symbol nkjp:manual="true" value="pun" xml:id="morph_1.9.2.1-msd"/>
          </f>
         </fs>
        </f>
        <f name="disamb">
         <fs feats="#an8003" type="tool_report">
          <f fVal="#morph_1.9.2.1-msd" name="choice"/>
          <f name="interpretation">
           <string>ustęp:brev:pun</string>
          </f>
         </fs>
        </f>
       </fs>
      </seg>
      <seg corresp="ann_segmentation.xml#segm_1.10-seg" xml:id="morph_1.10-seg">
       <fs type="morph">
        <f name="orth">
         <string>.</string>
        </f>
        <!-- . [32,1] -->
        <f name="nps">
         <binary value="true"/>
        </f>
        <f name="interps">
         <fs type="lex" xml:id="morph_1.10.1-lex">
          <f name="base">
           <string>.</string>
          </f>
          <f name="ctag">
           <symbol value="interp"/>
          </f>
          <f name="msd">
           <symbol value="" xml:id="morph_1.10.1.1-msd"/>
          </f>
         </fs>
        </f>
        <f name="disamb">
         <fs feats="#an8003" type="tool_report">
          <f fVal="#morph_1.10.1.1-msd" name="choice"/>
          <f name="interpretation">
           <string>.:interp</string>
          </f>
         </fs>
        </f>
       </fs>
      </seg>
      <seg corresp="ann_segmentation.xml#segm_1.11-seg" xml:id="morph_1.11-seg">
       <fs type="morph">
        <f name="orth">
         <string>1</string>
        </f>
        <!-- 1 [34,1] -->
        <f name="interps">
         <fs type="lex" xml:id="morph_1.11.1-lex">
          <f name="base">
           <string/>
          </f>
          <f name="ctag">
           <symbol value="ign"/>
          </f>
          <f name="msd">
           <symbol value="" xml:id="morph_1.11.1.1-msd"/>
          </f>
         </fs>
         <fs type="lex" xml:id="morph_1.11.2-lex">
          <f name="base">
           <string>1</string>
          </f>
          <f name="ctag">
           <symbol value="adj"/>
          </f>
          <f name="msd">
           <symbol nkjp:manual="true" value="sg:loc:m3:pos" xml:id="morph_1.11.2.1-msd"/>
          </f>
         </fs>
        </f>
        <f name="disamb">
         <fs feats="#an8003" type="tool_report">
          <f fVal="#morph_1.11.2.1-msd" name="choice"/>
          <f name="interpretation">
           <string>1:adj:sg:loc:m3:pos</string>
          </f>
         </fs>
        </f>
       </fs>
      </seg>
      <seg corresp="ann_segmentation.xml#segm_1.12-seg" xml:id="morph_1.12-seg">
       <fs type="morph">
        <f name="orth">
         <string>,</string>
        </f>
        <!-- , [35,1] -->
        <f name="nps">
         <binary value="true"/>
        </f>
        <f name="interps">
         <fs type="lex" xml:id="morph_1.12.1-lex">
          <f name="base">
           <string>,</string>
          </f>
          <f name="ctag">
           <symbol value="interp"/>
          </f>
          <f name="msd">
           <symbol value="" xml:id="morph_1.12.1.1-msd"/>
          </f>
         </fs>
        </f>
        <f name="disamb">
         <fs feats="#an8003" type="tool_report">
          <f fVal="#morph_1.12.1.1-msd" name="choice"/>
          <f name="interpretation">
           <string>,:interp</string>
          </f>
         </fs>
        </f>
       </fs>
      </seg>
      <seg corresp="ann_segmentation.xml#segm_1.13-seg" xml:id="morph_1.13-seg">
       <fs type="morph">
        <f name="orth">
         <string>doręcza</string>
        </f>
        <!-- doręcza [37,7] -->
        <f name="interps">
         <fs type="lex" xml:id="morph_1.13.1-lex">
          <f name="base">
           <string>doręczać</string>
          </f>
          <f name="ctag">
           <symbol value="fin"/>
          </f>
          <f name="msd">
           <symbol value="sg:ter:imperf" xml:id="morph_1.13.1.1-msd"/>
          </f>
         </fs>
         <fs type="lex" xml:id="morph_1.13.2-lex">
          <f name="base">
           <string>doręcze</string>
          </f>
          <f name="ctag">
           <symbol value="subst"/>
          </f>
          <f name="msd">
           <vAlt>
            <symbol value="sg:gen:n" xml:id="morph_1.13.2.1-msd"/>
            <symbol value="pl:nom:n" xml:id="morph_1.13.2.2-msd"/>
            <symbol value="pl:acc:n" xml:id="morph_1.13.2.3-msd"/>
            <symbol value="pl:voc:n" xml:id="morph_1.13.2.4-msd"/>
           </vAlt>
          </f>
         </fs>
        </f>
        <f name="disamb">
         <fs feats="#an8003" type="tool_report">
          <f fVal="#morph_1.13.1.1-msd" name="choice"/>
          <f name="interpretation">
           <string>doręczać:fin:sg:ter:imperf</string>
          </f>
         </fs>
        </f>
       </fs>
      </seg>
      <seg corresp="ann_segmentation.xml#segm_1.14-seg" xml:id="morph_1.14-seg">
       <fs type="morph">
        <f name="orth">
         <string>się</string>
        </f>
        <!-- się [45,3] -->
        <f name="interps">
         <fs type="lex" xml:id="morph_1.14.1-lex">
          <f name="base">
           <string>się</string>
          </f>
          <f name="ctag">
           <symbol value="qub"/>
          </f>
          <f name="msd">
           <symbol value="" xml:id="morph_1.14.1.1-msd"/>
          </f>
         </fs>
        </f>
        <f name="disamb">
         <fs feats="#an8003" type="tool_report">
          <f fVal="#morph_1.14.1.1-msd" name="choice"/>
          <f name="interpretation">
           <string>się:qub</string>
          </f>
         </fs>
        </f>
       </fs>
      </seg>
      <seg corresp="ann_segmentation.xml#segm_1.15-seg" xml:id="morph_1.15-seg">
       <fs type="morph">
        <f name="orth">
         <string>na</string>
        </f>
        <!-- na [49,2] -->
        <f name="interps">
         <fs type="lex" xml:id="morph_1.15.1-lex">
          <f name="base">
           <string>na</string>
          </f>
          <f name="ctag">
           <symbol value="interj"/>
          </f>
          <f name="msd">
           <symbol value="" xml:id="morph_1.15.1.1-msd"/>
          </f>
         </fs>
         <fs type="lex" xml:id="morph_1.15.2-lex">
          <f name="base">
           <string>na</string>
          </f>
          <f name="ctag">
           <symbol value="prep"/>
          </f>
          <f name="msd">
           <vAlt>
            <symbol value="acc" xml:id="morph_1.15.2.1-msd"/>
            <symbol value="loc" xml:id="morph_1.15.2.2-msd"/>
           </vAlt>
          </f>
         </fs>
        </f>
        <f name="disamb">
         <fs feats="#an8003" type="tool_report">
          <f fVal="#morph_1.15.2.1-msd" name="choice"/>
          <f name="interpretation">
           <string>na:prep:acc</string>
          </f>
         </fs>
        </f>
       </fs>
      </seg>
      <seg corresp="ann_segmentation.xml#segm_1.16-seg" xml:id="morph_1.16-seg">
       <fs type="morph">
        <f name="orth">
         <string>czternaście</string>
        </f>
        <!-- czternaście [52,11] -->
        <f name="interps">
         <fs type="lex" xml:id="morph_1.16.1-lex">
          <f name="base">
           <string>czternaście</string>
          </f>
          <f name="ctag">
           <symbol value="num"/>
          </f>
          <f name="msd">
           <vAlt>
            <symbol value="pl:nom:m2:rec" xml:id="morph_1.16.1.1-msd"/>
            <symbol value="pl:nom:m3:rec" xml:id="morph_1.16.1.2-msd"/>
            <symbol value="pl:nom:f:rec" xml:id="morph_1.16.1.3-msd"/>
            <symbol value="pl:nom:n:rec" xml:id="morph_1.16.1.4-msd"/>
            <symbol value="pl:acc:m2:rec" xml:id="morph_1.16.1.5-msd"/>
            <symbol value="pl:acc:m3:rec" xml:id="morph_1.16.1.6-msd"/>
            <symbol value="pl:acc:f:rec" xml:id="morph_1.16.1.7-msd"/>
            <symbol value="pl:acc:n:rec" xml:id="morph_1.16.1.8-msd"/>
            <symbol value="pl:voc:m2:rec" xml:id="morph_1.16.1.9-msd"/>
            <symbol value="pl:voc:m3:rec" xml:id="morph_1.16.1.10-msd"/>
            <symbol value="pl:voc:f:rec" xml:id="morph_1.16.1.11-msd"/>
            <symbol value="pl:voc:n:rec" xml:id="morph_1.16.1.12-msd"/>
           </vAlt>
          </f>
         </fs>
        </f>
        <f name="disamb">
         <fs feats="#an8003" type="tool_report">
          <f fVal="#morph_1.16.1.6-msd" name="choice"/>
          <f name="interpretation">
           <string>czternaście:num:pl:acc:m3:rec</string>
          </f>
         </fs>
        </f>
       </fs>
      </seg>
      <seg corresp="ann_segmentation.xml#segm_1.17-seg" xml:id="morph_1.17-seg">
       <fs type="morph">
        <f name="orth">
         <string>dni</string>
        </f>
        <!-- dni [64,3] -->
        <f name="interps">
         <fs type="lex" xml:id="morph_1.17.1-lex">
          <f name="base">
           <string>dni</string>
          </f>
          <f name="ctag">
           <symbol value="subst"/>
          </f>
          <f name="msd">
           <vAlt>
            <symbol value="pl:nom:n" xml:id="morph_1.17.1.1-msd"/>
            <symbol value="pl:gen:n" xml:id="morph_1.17.1.2-msd"/>
            <symbol value="pl:acc:n" xml:id="morph_1.17.1.3-msd"/>
            <symbol value="pl:voc:n" xml:id="morph_1.17.1.4-msd"/>
           </vAlt>
          </f>
         </fs>
         <fs type="lex" xml:id="morph_1.17.2-lex">
          <f name="base">
           <string>dzień</string>
          </f>
          <f name="ctag">
           <symbol value="subst"/>
          </f>
          <f name="msd">
           <vAlt>
            <symbol value="pl:nom:m3" xml:id="morph_1.17.2.1-msd"/>
            <symbol value="pl:gen:m3" xml:id="morph_1.17.2.2-msd"/>
            <symbol value="pl:acc:m3" xml:id="morph_1.17.2.3-msd"/>
            <symbol value="pl:voc:m3" xml:id="morph_1.17.2.4-msd"/>
           </vAlt>
          </f>
         </fs>
        </f>
        <f name="disamb">
         <fs feats="#an8003" type="tool_report">
          <f fVal="#morph_1.17.2.2-msd" name="choice"/>
          <f name="interpretation">
           <string>dzień:subst:pl:gen:m3</string>
          </f>
         </fs>
        </f>
       </fs>
      </seg>
      <seg corresp="ann_segmentation.xml#segm_1.18-seg" xml:id="morph_1.18-seg">
       <fs type="morph">
        <f name="orth">
         <string>przed</string>
        </f>
        <!-- przed [68,5] -->
        <f name="interps">
         <fs type="lex" xml:id="morph_1.18.1-lex">
          <f name="base">
           <string>przed</string>
          </f>
          <f name="ctag">
           <symbol value="prep"/>
          </f>
          <f name="msd">
           <vAlt>
            <symbol value="acc:nwok" xml:id="morph_1.18.1.1-msd"/>
            <symbol value="inst:nwok" xml:id="morph_1.18.1.2-msd"/>
           </vAlt>
          </f>
         </fs>
        </f>
        <f name="disamb">
         <fs feats="#an8003" type="tool_report">
          <f fVal="#morph_1.18.1.2-msd" name="choice"/>
          <f name="interpretation">
           <string>przed:prep:inst:nwok</string>
          </f>
         </fs>
        </f>
       </fs>
      </seg>
      <seg corresp="ann_segmentation.xml#segm_1.19-seg" xml:id="morph_1.19-seg">
       <fs type="morph">
        <f name="orth">
         <string>terminem</string>
        </f>
        <!-- terminem [74,8] -->
        <f name="interps">
         <fs type="lex" xml:id="morph_1.19.1-lex">
          <f name="base">
           <string>termin</string>
          </f>
          <f name="ctag">
           <symbol value="subst"/>
          </f>
          <f name="msd">
           <symbol value="sg:inst:m3" xml:id="morph_1.19.1.1-msd"/>
          </f>
         </fs>
        </f>
        <f name="disamb">
         <fs feats="#an8003" type="tool_report">
          <f fVal="#morph_1.19.1.1-msd" name="choice"/>
          <f name="interpretation">
           <string>termin:subst:sg:inst:m3</string>
          </f>
         </fs>
        </f>
       </fs>
      </seg>
      <seg corresp="ann_segmentation.xml#segm_1.20-seg" xml:id="morph_1.20-seg">
       <fs type="morph">
        <f name="orth">
         <string>wykonania</string>
        </f>
        <!-- wykonania [83,9] -->
        <f name="interps">
         <fs type="lex" xml:id="morph_1.20.1-lex">
          <f name="base">
           <string>wykonanie</string>
          </f>
          <f name="ctag">
           <symbol value="subst"/>
          </f>
          <f name="msd">
           <vAlt>
            <symbol value="sg:gen:n" xml:id="morph_1.20.1.1-msd"/>
            <symbol value="pl:nom:n" xml:id="morph_1.20.1.2-msd"/>
            <symbol value="pl:acc:n" xml:id="morph_1.20.1.3-msd"/>
            <symbol value="pl:voc:n" xml:id="morph_1.20.1.4-msd"/>
           </vAlt>
          </f>
         </fs>
         <fs type="lex" xml:id="morph_1.20.2-lex">
          <f name="base">
           <string>wykonać</string>
          </f>
          <f name="ctag">
           <symbol value="ger"/>
          </f>
          <f name="msd">
           <symbol value="sg:gen:n:perf:aff" xml:id="morph_1.20.2.1-msd"/>
          </f>
         </fs>
        </f>
        <f name="disamb">
         <fs feats="#an8003" type="tool_report">
          <f fVal="#morph_1.20.2.1-msd" name="choice"/>
          <f name="interpretation">
           <string>wykonać:ger:sg:gen:n:perf:aff</string>
          </f>
         </fs>
        </f>
       </fs>
      </seg>
      <seg corresp="ann_segmentation.xml#segm_1.21-seg" xml:id="morph_1.21-seg">
       <fs type="morph">
        <f name="orth">
         <string>świadczenia</string>
        </f>
        <!-- świadczenia [93,11] -->
        <f name="interps">
         <fs type="lex" xml:id="morph_1.21.1-lex">
          <f name="base">
           <string>świadczenie</string>
          </f>
          <f name="ctag">
           <symbol value="subst"/>
          </f>
          <f name="msd">
           <vAlt>
            <symbol value="sg:gen:n" xml:id="morph_1.21.1.1-msd"/>
            <symbol value="pl:nom:n" xml:id="morph_1.21.1.2-msd"/>
            <symbol value="pl:acc:n" xml:id="morph_1.21.1.3-msd"/>
            <symbol value="pl:voc:n" xml:id="morph_1.21.1.4-msd"/>
           </vAlt>
          </f>
         </fs>
         <fs type="lex" xml:id="morph_1.21.2-lex">
          <f name="base">
           <string>świadczyć</string>
          </f>
          <f name="ctag">
           <symbol value="ger"/>
          </f>
          <f name="msd">
           <symbol value="sg:gen:n:imperf:aff" xml:id="morph_1.21.2.1-msd"/>
          </f>
         </fs>
        </f>
        <f name="disamb">
         <fs feats="#an8003" type="tool_report">
          <f fVal="#morph_1.21.1.1-msd" name="choice"/>
          <f name="interpretation">
           <string>świadczenie:subst:sg:gen:n</string>
          </f>
         </fs>
        </f>
       </fs>
      </seg>
      <seg corresp="ann_segmentation.xml#segm_1.22-seg" xml:id="morph_1.22-seg">
       <fs type="morph">
        <f name="orth">
         <string>,</string>
        </f>
        <!-- , [104,1] -->
        <f name="nps">
         <binary value="true"/>
        </f>
        <f name="interps">
         <fs type="lex" xml:id="morph_1.22.1-lex">
          <f name="base">
           <string>,</string>
          </f>
          <f name="ctag">
           <symbol value="interp"/>
          </f>
          <f name="msd">
           <symbol value="" xml:id="morph_1.22.1.1-msd"/>
          </f>
         </fs>
        </f>
        <f name="disamb">
         <fs feats="#an8003" type="tool_report">
          <f fVal="#morph_1.22.1.1-msd" name="choice"/>
          <f name="interpretation">
           <string>,:interp</string>
          </f>
         </fs>
        </f>
       </fs>
      </seg>
      <seg corresp="ann_segmentation.xml#segm_1.23-seg" xml:id="morph_1.23-seg">
       <fs type="morph">
        <f name="orth">
         <string>z</string>
        </f>
        <!-- z [106,1] -->
        <f name="interps">
         <fs type="lex" xml:id="morph_1.23.1-lex">
          <f name="base">
           <string>z</string>
          </f>
          <f name="ctag">
           <symbol value="prep"/>
          </f>
          <f name="msd">
           <vAlt>
            <symbol value="gen:nwok" xml:id="morph_1.23.1.1-msd"/>
            <symbol value="acc:nwok" xml:id="morph_1.23.1.2-msd"/>
            <symbol value="inst:nwok" xml:id="morph_1.23.1.3-msd"/>
           </vAlt>
          </f>
         </fs>
         <fs type="lex" xml:id="morph_1.23.2-lex">
          <f name="base">
           <string>z</string>
          </f>
          <f name="ctag">
           <symbol value="qub"/>
          </f>
          <f name="msd">
           <symbol value="" xml:id="morph_1.23.2.1-msd"/>
          </f>
         </fs>
         <fs type="lex" xml:id="morph_1.23.3-lex">
          <f name="base">
           <string>zeszyt</string>
          </f>
          <f name="ctag">
           <symbol value="brev"/>
          </f>
          <f name="msd">
           <symbol value="pun" xml:id="morph_1.23.3.1-msd"/>
          </f>
         </fs>
        </f>
        <f name="disamb">
         <fs feats="#an8003" type="tool_report">
          <f fVal="#morph_1.23.1.3-msd" name="choice"/>
          <f name="interpretation">
           <string>z:prep:inst:nwok</string>
          </f>
         </fs>
        </f>
       </fs>
      </seg>
      <seg corresp="ann_segmentation.xml#segm_1.24-seg" xml:id="morph_1.24-seg">
       <fs type="morph">
        <f name="orth">
         <string>wyjątkiem</string>
        </f>
        <!-- wyjątkiem [108,9] -->
        <f name="interps">
         <fs type="lex" xml:id="morph_1.24.1-lex">
          <f name="base">
           <string>wyjątek</string>
          </f>
          <f name="ctag">
           <symbol value="subst"/>
          </f>
          <f name="msd">
           <symbol value="sg:inst:m3" xml:id="morph_1.24.1.1-msd"/>
          </f>
         </fs>
        </f>
        <f name="disamb">
         <fs feats="#an8003" type="tool_report">
          <f fVal="#morph_1.24.1.1-msd" name="choice"/>
          <f name="interpretation">
           <string>wyjątek:subst:sg:inst:m3</string>
          </f>
         </fs>
        </f>
       </fs>
      </seg>
      <seg corresp="ann_segmentation.xml#segm_1.25-seg" xml:id="morph_1.25-seg">
       <fs type="morph">
        <f name="orth">
         <string>przypadków</string>
        </f>
        <!-- przypadków [118,10] -->
        <f name="interps">
         <fs type="lex" xml:id="morph_1.25.1-lex">
          <f name="base">
           <string>przypadek</string>
          </f>
          <f name="ctag">
           <symbol value="subst"/>
          </f>
          <f name="msd">
           <symbol value="pl:gen:m3" xml:id="morph_1.25.1.1-msd"/>
          </f>
         </fs>
        </f>
        <f name="disamb">
         <fs feats="#an8003" type="tool_report">
          <f fVal="#morph_1.25.1.1-msd" name="choice"/>
          <f name="interpretation">
           <string>przypadek:subst:pl:gen:m3</string>
          </f>
         </fs>
        </f>
       </fs>
      </seg>
      <seg corresp="ann_segmentation.xml#segm_1.26-seg" xml:id="morph_1.26-seg">
       <fs type="morph">
        <f name="orth">
         <string>,</string>
        </f>
        <!-- , [128,1] -->
        <f name="nps">
         <binary value="true"/>
        </f>
        <f name="interps">
         <fs type="lex" xml:id="morph_1.26.1-lex">
          <f name="base">
           <string>,</string>
          </f>
          <f name="ctag">
           <symbol value="interp"/>
          </f>
          <f name="msd">
           <symbol value="" xml:id="morph_1.26.1.1-msd"/>
          </f>
         </fs>
        </f>
        <f name="disamb">
         <fs feats="#an8003" type="tool_report">
          <f fVal="#morph_1.26.1.1-msd" name="choice"/>
          <f name="interpretation">
           <string>,:interp</string>
          </f>
         </fs>
        </f>
       </fs>
      </seg>
      <seg corresp="ann_segmentation.xml#segm_1.27-seg" xml:id="morph_1.27-seg">
       <fs type="morph">
        <f name="orth">
         <string>w</string>
        </f>
        <!-- w [130,1] -->
        <f name="interps">
         <fs type="lex" xml:id="morph_1.27.1-lex">
          <f name="base">
           <string>w</string>
          </f>
          <f name="ctag">
           <symbol value="prep"/>
          </f>
          <f name="msd">
           <vAlt>
            <symbol value="acc:nwok" xml:id="morph_1.27.1.1-msd"/>
            <symbol value="loc:nwok" xml:id="morph_1.27.1.2-msd"/>
           </vAlt>
          </f>
         </fs>
         <fs type="lex" xml:id="morph_1.27.2-lex">
          <f name="base">
           <string>wiek</string>
          </f>
          <f name="ctag">
           <symbol value="brev"/>
          </f>
          <f name="msd">
           <symbol value="pun" xml:id="morph_1.27.2.1-msd"/>
          </f>
         </fs>
         <fs type="lex" xml:id="morph_1.27.3-lex">
          <f name="base">
           <string>wielki</string>
          </f>
          <f name="ctag">
           <symbol value="brev"/>
          </f>
          <f name="msd">
           <symbol value="pun" xml:id="morph_1.27.3.1-msd"/>
          </f>
         </fs>
         <fs type="lex" xml:id="morph_1.27.4-lex">
          <f name="base">
           <string>wiersz</string>
          </f>
          <f name="ctag">
           <symbol value="brev"/>
          </f>
          <f name="msd">
           <symbol value="pun" xml:id="morph_1.27.4.1-msd"/>
          </f>
         </fs>
         <fs type="lex" xml:id="morph_1.27.5-lex">
          <f name="base">
           <string>wieś</string>
          </f>
          <f name="ctag">
           <symbol value="brev"/>
          </f>
          <f name="msd">
           <symbol value="pun" xml:id="morph_1.27.5.1-msd"/>
          </f>
         </fs>
         <fs type="lex" xml:id="morph_1.27.6-lex">
          <f name="base">
           <string>wyspa</string>
          </f>
          <f name="ctag">
           <symbol value="brev"/>
          </f>
          <f name="msd">
           <symbol value="pun" xml:id="morph_1.27.6.1-msd"/>
          </f>
         </fs>
        </f>
        <f name="disamb">
         <fs feats="#an8003" type="tool_report">
          <f fVal="#morph_1.27.1.2-msd" name="choice"/>
          <f name="interpretation">
           <string>w:prep:loc:nwok</string>
          </f>
         </fs>
        </f>
       </fs>
      </seg>
      <seg corresp="ann_segmentation.xml#segm_1.28-seg" xml:id="morph_1.28-seg">
       <fs type="morph">
        <f name="orth">
         <string>których</string>
        </f>
        <!-- których [132,7] -->
        <f name="interps">
         <fs type="lex" xml:id="morph_1.28.1-lex">
          <f name="base">
           <string>który</string>
          </f>
          <f name="ctag">
           <symbol value="adj"/>
          </f>
          <f name="msd">
           <vAlt>
            <symbol value="pl:gen:m1:pos" xml:id="morph_1.28.1.1-msd"/>
            <symbol value="pl:gen:m2:pos" xml:id="morph_1.28.1.2-msd"/>
            <symbol value="pl:gen:m3:pos" xml:id="morph_1.28.1.3-msd"/>
            <symbol value="pl:gen:f:pos" xml:id="morph_1.28.1.4-msd"/>
            <symbol value="pl:gen:n:pos" xml:id="morph_1.28.1.5-msd"/>
            <symbol value="pl:loc:m1:pos" xml:id="morph_1.28.1.6-msd"/>
            <symbol value="pl:loc:m2:pos" xml:id="morph_1.28.1.7-msd"/>
            <symbol value="pl:loc:m3:pos" xml:id="morph_1.28.1.8-msd"/>
            <symbol value="pl:loc:f:pos" xml:id="morph_1.28.1.9-msd"/>
            <symbol value="pl:loc:n:pos" xml:id="morph_1.28.1.10-msd"/>
            <symbol value="pl:acc:m1:pos" xml:id="morph_1.28.1.11-msd"/>
           </vAlt>
          </f>
         </fs>
        </f>
        <f name="disamb">
         <fs feats="#an8003" type="tool_report">
          <f fVal="#morph_1.28.1.8-msd" name="choice"/>
          <f name="interpretation">
           <string>który:adj:pl:loc:m3:pos</string>
          </f>
         </fs>
        </f>
       </fs>
      </seg>
      <seg corresp="ann_segmentation.xml#segm_1.29-seg" xml:id="morph_1.29-seg">
       <fs type="morph">
        <f name="orth">
         <string>wykonanie</string>
        </f>
        <!-- wykonanie [140,9] -->
        <f name="interps">
         <fs type="lex" xml:id="morph_1.29.1-lex">
          <f name="base">
           <string>wykonanie</string>
          </f>
          <f name="ctag">
           <symbol value="subst"/>
          </f>
          <f name="msd">
           <vAlt>
            <symbol value="sg:nom:n" xml:id="morph_1.29.1.1-msd"/>
            <symbol value="sg:acc:n" xml:id="morph_1.29.1.2-msd"/>
            <symbol value="sg:voc:n" xml:id="morph_1.29.1.3-msd"/>
           </vAlt>
          </f>
         </fs>
         <fs type="lex" xml:id="morph_1.29.2-lex">
          <f name="base">
           <string>wykonać</string>
          </f>
          <f name="ctag">
           <symbol value="ger"/>
          </f>
          <f name="msd">
           <vAlt>
            <symbol value="sg:nom:n:perf:aff" xml:id="morph_1.29.2.1-msd"/>
            <symbol value="sg:acc:n:perf:aff" xml:id="morph_1.29.2.2-msd"/>
           </vAlt>
          </f>
         </fs>
        </f>
        <f name="disamb">
         <fs feats="#an8003" type="tool_report">
          <f fVal="#morph_1.29.2.1-msd" name="choice"/>
          <f name="interpretation">
           <string>wykonać:ger:sg:nom:n:perf:aff</string>
          </f>
         </fs>
        </f>
       </fs>
      </seg>
      <seg corresp="ann_segmentation.xml#segm_1.30-seg" xml:id="morph_1.30-seg">
       <fs type="morph">
        <f name="orth">
         <string>świadczenia</string>
        </f>
        <!-- świadczenia [150,11] -->
        <f name="interps">
         <fs type="lex" xml:id="morph_1.30.1-lex">
          <f name="base">
           <string>świadczenie</string>
          </f>
          <f name="ctag">
           <symbol value="subst"/>
          </f>
          <f name="msd">
           <vAlt>
            <symbol value="sg:gen:n" xml:id="morph_1.30.1.1-msd"/>
            <symbol value="pl:nom:n" xml:id="morph_1.30.1.2-msd"/>
            <symbol value="pl:acc:n" xml:id="morph_1.30.1.3-msd"/>
            <symbol value="pl:voc:n" xml:id="morph_1.30.1.4-msd"/>
           </vAlt>
          </f>
         </fs>
         <fs type="lex" xml:id="morph_1.30.2-lex">
          <f name="base">
           <string>świadczyć</string>
          </f>
          <f name="ctag">
           <symbol value="ger"/>
          </f>
          <f name="msd">
           <symbol value="sg:gen:n:imperf:aff" xml:id="morph_1.30.2.1-msd"/>
          </f>
         </fs>
        </f>
        <f name="disamb">
         <fs feats="#an8003" type="tool_report">
          <f fVal="#morph_1.30.1.1-msd" name="choice"/>
          <f name="interpretation">
           <string>świadczenie:subst:sg:gen:n</string>
          </f>
         </fs>
        </f>
       </fs>
      </seg>
      <seg corresp="ann_segmentation.xml#segm_1.31-seg" xml:id="morph_1.31-seg">
       <fs type="morph">
        <f name="orth">
         <string>następuje</string>
        </f>
        <!-- następuje [162,9] -->
        <f name="interps">
         <fs type="lex" xml:id="morph_1.31.1-lex">
          <f name="base">
           <string>następować</string>
          </f>
          <f name="ctag">
           <symbol value="fin"/>
          </f>
          <f name="msd">
           <symbol value="sg:ter:imperf" xml:id="morph_1.31.1.1-msd"/>
          </f>
         </fs>
        </f>
        <f name="disamb">
         <fs feats="#an8003" type="tool_report">
          <f fVal="#morph_1.31.1.1-msd" name="choice"/>
          <f name="interpretation">
           <string>następować:fin:sg:ter:imperf</string>
          </f>
         </fs>
        </f>
       </fs>
      </seg>
      <seg corresp="ann_segmentation.xml#segm_1.32-seg" xml:id="morph_1.32-seg">
       <fs type="morph">
        <f name="orth">
         <string>w</string>
        </f>
        <!-- w [172,1] -->
        <f name="interps">
         <fs type="lex" xml:id="morph_1.32.1-lex">
          <f name="base">
           <string>w</string>
          </f>
          <f name="ctag">
           <symbol value="prep"/>
          </f>
          <f name="msd">
           <vAlt>
            <symbol value="acc:nwok" xml:id="morph_1.32.1.1-msd"/>
            <symbol value="loc:nwok" xml:id="morph_1.32.1.2-msd"/>
           </vAlt>
          </f>
         </fs>
         <fs type="lex" xml:id="morph_1.32.2-lex">
          <f name="base">
           <string>wiek</string>
          </f>
          <f name="ctag">
           <symbol value="brev"/>
          </f>
          <f name="msd">
           <symbol value="pun" xml:id="morph_1.32.2.1-msd"/>
          </f>
         </fs>
         <fs type="lex" xml:id="morph_1.32.3-lex">
          <f name="base">
           <string>wielki</string>
          </f>
          <f name="ctag">
           <symbol value="brev"/>
          </f>
          <f name="msd">
           <symbol value="pun" xml:id="morph_1.32.3.1-msd"/>
          </f>
         </fs>
         <fs type="lex" xml:id="morph_1.32.4-lex">
          <f name="base">
           <string>wiersz</string>
          </f>
          <f name="ctag">
           <symbol value="brev"/>
          </f>
          <f name="msd">
           <symbol value="pun" xml:id="morph_1.32.4.1-msd"/>
          </f>
         </fs>
         <fs type="lex" xml:id="morph_1.32.5-lex">
          <f name="base">
           <string>wieś</string>
          </f>
          <f name="ctag">
           <symbol value="brev"/>
          </f>
          <f name="msd">
           <symbol value="pun" xml:id="morph_1.32.5.1-msd"/>
          </f>
         </fs>
         <fs type="lex" xml:id="morph_1.32.6-lex">
          <f name="base">
           <string>wyspa</string>
          </f>
          <f name="ctag">
           <symbol value="brev"/>
          </f>
          <f name="msd">
           <symbol value="pun" xml:id="morph_1.32.6.1-msd"/>
          </f>
         </fs>
        </f>
        <f name="disamb">
         <fs feats="#an8003" type="tool_report">
          <f fVal="#morph_1.32.1.2-msd" name="choice"/>
          <f name="interpretation">
           <string>w:prep:loc:nwok</string>
          </f>
         </fs>
        </f>
       </fs>
      </seg>
      <seg corresp="ann_segmentation.xml#segm_1.33-seg" xml:id="morph_1.33-seg">
       <fs type="morph">
        <f name="orth">
         <string>celu</string>
        </f>
        <!-- celu [174,4] -->
        <f name="interps">
         <fs type="lex" xml:id="morph_1.33.1-lex">
          <f name="base">
           <string>Cela</string>
          </f>
          <f name="ctag">
           <symbol value="subst"/>
          </f>
          <f name="msd">
           <symbol value="sg:voc:f" xml:id="morph_1.33.1.1-msd"/>
          </f>
         </fs>
         <fs type="lex" xml:id="morph_1.33.2-lex">
          <f name="base">
           <string>cel</string>
          </f>
          <f name="ctag">
           <symbol value="subst"/>
          </f>
          <f name="msd">
           <vAlt>
            <symbol value="sg:gen:m3" xml:id="morph_1.33.2.1-msd"/>
            <symbol value="sg:loc:m3" xml:id="morph_1.33.2.2-msd"/>
            <symbol value="sg:voc:m3" xml:id="morph_1.33.2.3-msd"/>
           </vAlt>
          </f>
         </fs>
        </f>
        <f name="disamb">
         <fs feats="#an8003" type="tool_report">
          <f fVal="#morph_1.33.2.2-msd" name="choice"/>
          <f name="interpretation">
           <string>cel:subst:sg:loc:m3</string>
          </f>
         </fs>
        </f>
       </fs>
      </seg>
      <seg corresp="ann_segmentation.xml#segm_1.34-seg" xml:id="morph_1.34-seg">
       <fs type="morph">
        <f name="orth">
         <string>sprawdzenia</string>
        </f>
        <!-- sprawdzenia [179,11] -->
        <f name="interps">
         <fs type="lex" xml:id="morph_1.34.1-lex">
          <f name="base">
           <string>sprawdzić</string>
          </f>
          <f name="ctag">
           <symbol value="ger"/>
          </f>
          <f name="msd">
           <symbol value="sg:gen:n:perf:aff" xml:id="morph_1.34.1.1-msd"/>
          </f>
         </fs>
        </f>
        <f name="disamb">
         <fs feats="#an8003" type="tool_report">
          <f fVal="#morph_1.34.1.1-msd" name="choice"/>
          <f name="interpretation">
           <string>sprawdzić:ger:sg:gen:n:perf:aff</string>
          </f>
         </fs>
        </f>
       </fs>
      </seg>
      <seg corresp="ann_segmentation.xml#segm_1.35-seg" xml:id="morph_1.35-seg">
       <fs type="morph">
        <f name="orth">
         <string>gotowości</string>
        </f>
        <!-- gotowości [191,9] -->
        <f name="interps">
         <fs type="lex" xml:id="morph_1.35.1-lex">
          <f name="base">
           <string>gotowość</string>
          </f>
          <f name="ctag">
           <symbol value="subst"/>
          </f>
          <f name="msd">
           <vAlt>
            <symbol value="sg:gen:f" xml:id="morph_1.35.1.1-msd"/>
            <symbol value="sg:dat:f" xml:id="morph_1.35.1.2-msd"/>
            <symbol value="sg:loc:f" xml:id="morph_1.35.1.3-msd"/>
            <symbol value="sg:voc:f" xml:id="morph_1.35.1.4-msd"/>
            <symbol value="pl:nom:f" xml:id="morph_1.35.1.5-msd"/>
            <symbol value="pl:gen:f" xml:id="morph_1.35.1.6-msd"/>
            <symbol value="pl:acc:f" xml:id="morph_1.35.1.7-msd"/>
            <symbol value="pl:voc:f" xml:id="morph_1.35.1.8-msd"/>
           </vAlt>
          </f>
         </fs>
        </f>
        <f name="disamb">
         <fs feats="#an8003" type="tool_report">
          <f fVal="#morph_1.35.1.1-msd" name="choice"/>
          <f name="interpretation">
           <string>gotowość:subst:sg:gen:f</string>
          </f>
         </fs>
        </f>
       </fs>
      </seg>
      <seg corresp="ann_segmentation.xml#segm_1.36-seg" xml:id="morph_1.36-seg">
       <fs type="morph">
        <f name="orth">
         <string>mobilizacyjnej</string>
        </f>
        <!-- mobilizacyjnej [201,14] -->
        <f name="interps">
         <fs type="lex" xml:id="morph_1.36.1-lex">
          <f name="base">
           <string>mobilizacyjny</string>
          </f>
          <f name="ctag">
           <symbol value="adj"/>
          </f>
          <f name="msd">
           <vAlt>
            <symbol value="sg:gen:f:pos" xml:id="morph_1.36.1.1-msd"/>
            <symbol value="sg:dat:f:pos" xml:id="morph_1.36.1.2-msd"/>
            <symbol value="sg:loc:f:pos" xml:id="morph_1.36.1.3-msd"/>
           </vAlt>
          </f>
         </fs>
        </f>
        <f name="disamb">
         <fs feats="#an8003" type="tool_report">
          <f fVal="#morph_1.36.1.1-msd" name="choice"/>
          <f name="interpretation">
           <string>mobilizacyjny:adj:sg:gen:f:pos</string>
          </f>
         </fs>
        </f>
       </fs>
      </seg>
      <seg corresp="ann_segmentation.xml#segm_1.37-seg" xml:id="morph_1.37-seg">
       <fs type="morph">
        <f name="orth">
         <string>Sił</string>
        </f>
        <!-- Sił [216,3] -->
        <f name="interps">
         <fs type="lex" xml:id="morph_1.37.1-lex">
          <f name="base">
           <string>siła</string>
          </f>
          <f name="ctag">
           <symbol value="subst"/>
          </f>
          <f name="msd">
           <symbol value="pl:gen:f" xml:id="morph_1.37.1.1-msd"/>
          </f>
         </fs>
         <fs type="lex" xml:id="morph_1.37.2-lex">
          <f name="base">
           <string>siły</string>
          </f>
          <f name="ctag">
           <symbol value="subst"/>
          </f>
          <f name="msd">
           <symbol value="pl:gen:n" xml:id="morph_1.37.2.1-msd"/>
          </f>
         </fs>
        </f>
        <f name="disamb">
         <fs feats="#an8003" type="tool_report">
          <f fVal="#morph_1.37.1.1-msd" name="choice"/>
          <f name="interpretation">
           <string>siła:subst:pl:gen:f</string>
          </f>
         </fs>
        </f>
       </fs>
      </seg>
      <seg corresp="ann_segmentation.xml#segm_1.38-seg" xml:id="morph_1.38-seg">
       <fs type="morph">
        <f name="orth">
         <string>Zbrojnych</string>
        </f>
        <!-- Zbrojnych [220,9] -->
        <f name="interps">
         <fs type="lex" xml:id="morph_1.38.1-lex">
          <f name="base">
           <string>zbrojny</string>
          </f>
          <f name="ctag">
           <symbol value="adj"/>
          </f>
          <f name="msd">
           <vAlt>
            <symbol value="pl:gen:m1:pos" xml:id="morph_1.38.1.1-msd"/>
            <symbol value="pl:gen:m2:pos" xml:id="morph_1.38.1.2-msd"/>
            <symbol value="pl:gen:m3:pos" xml:id="morph_1.38.1.3-msd"/>
            <symbol value="pl:gen:f:pos" xml:id="morph_1.38.1.4-msd"/>
            <symbol value="pl:gen:n:pos" xml:id="morph_1.38.1.5-msd"/>
            <symbol value="pl:loc:m1:pos" xml:id="morph_1.38.1.6-msd"/>
            <symbol value="pl:loc:m2:pos" xml:id="morph_1.38.1.7-msd"/>
            <symbol value="pl:loc:m3:pos" xml:id="morph_1.38.1.8-msd"/>
            <symbol value="pl:loc:f:pos" xml:id="morph_1.38.1.9-msd"/>
            <symbol value="pl:loc:n:pos" xml:id="morph_1.38.1.10-msd"/>
            <symbol value="pl:acc:m1:pos" xml:id="morph_1.38.1.11-msd"/>
           </vAlt>
          </f>
         </fs>
        </f>
        <f name="disamb">
         <fs feats="#an8003" type="tool_report">
          <f fVal="#morph_1.38.1.4-msd" name="choice"/>
          <f name="interpretation">
           <string>zbrojny:adj:pl:gen:f:pos</string>
          </f>
         </fs>
        </f>
       </fs>
      </seg>
      <seg corresp="ann_segmentation.xml#segm_1.39-seg" xml:id="morph_1.39-seg">
       <fs type="morph">
        <f name="orth">
         <string>.</string>
        </f>
        <!-- . [229,1] -->
        <f name="nps">
         <binary value="true"/>
        </f>
        <f name="interps">
         <fs type="lex" xml:id="morph_1.39.1-lex">
          <f name="base">
           <string>.</string>
          </f>
          <f name="ctag">
           <symbol value="interp"/>
          </f>
          <f name="msd">
           <symbol value="" xml:id="morph_1.39.1.1-msd"/>
          </f>
         </fs>
        </f>
        <f name="disamb">
         <fs feats="#an8003" type="tool_report">
          <f fVal="#morph_1.39.1.1-msd" name="choice"/>
          <f name="interpretation">
           <string>.:interp</string>
          </f>
         </fs>
        </f>
       </fs>
      </seg>
     </s>
    </p>
   </body>
  </text>
 </TEI>
</teiCorpus>
""".lstrip()



import torch
import torch.nn as nn
import os
import logging
import math
from torch.nn.utils.rnn import pad_sequence, pack_padded_sequence, pad_packed_sequence
from stanza.models.common.char_model import CharacterModel, CharacterLanguageModel
from typing import List, Tuple

from stanza.models.common.vocab import UNK_ID
from stanza.models.lemma_classifier import utils
from stanza.models.lemma_classifier.base_model import LemmaClassifier
from stanza.models.lemma_classifier.constants import ModelType

logger = logging.getLogger('stanza.lemmaclassifier')

class LemmaClassifierLSTM(LemmaClassifier):
    """
    Model architecture:
        Extracts word embeddings over the sentence, passes embeddings into a bi-LSTM to get a sentence encoding.
        From the LSTM output, we get the embedding of the specific token that we classify on. That embedding
        is fed into an MLP for classification.
    """
    def __init__(self, model_args, output_dim, pt_embedding, label_decoder, upos_to_id, known_words, target_words, target_upos,
                 use_charlm=False, charlm_forward_file=None, charlm_backward_file=None):
        """
        Args:
            vocab_size (int): Size of the vocab being used (if custom vocab)
            output_dim (int): Size of output vector from MLP layer
            upos_to_id (Mapping[str, int]): A dictionary mapping UPOS tag strings to their respective IDs
            pt_embedding (Pretrain): pretrained embeddings
            known_words (list(str)): Words which are in the training data
            target_words (set(str)): a set of the words which might need lemmatization
            use_charlm (bool): Whether or not to use the charlm embeddings
            charlm_forward_file (str): The path to the forward pass model for the character language model
            charlm_backward_file (str): The path to the forward pass model for the character language model.

        Kwargs:
            upos_emb_dim (int): The size of the UPOS tag embeddings
            num_heads (int): The number of heads to use for attention. If there are more than 0 heads, attention will be used instead of the LSTM.

        Raises:
            FileNotFoundError: if the forward or backward charlm file cannot be found.
        """
        super(LemmaClassifierLSTM, self).__init__(label_decoder, target_words, target_upos)
        self.model_args = model_args

        self.hidden_dim = model_args['hidden_dim']
        self.input_size = 0
        self.num_heads = self.model_args['num_heads']

        emb_matrix = pt_embedding.emb
        self.add_unsaved_module("embeddings", nn.Embedding.from_pretrained(emb_matrix, freeze=True))
        self.vocab_map = { word.replace('\xa0', ' '): i for i, word in enumerate(pt_embedding.vocab) }
        self.vocab_size = emb_matrix.shape[0]
        self.embedding_dim = emb_matrix.shape[1]

        self.known_words = known_words
        self.known_word_map = {word: idx for idx, word in enumerate(known_words)}
        self.delta_embedding = nn.Embedding(num_embeddings=len(known_words)+1,
                                            embedding_dim=self.embedding_dim,
                                            padding_idx=0)
        nn.init.normal_(self.delta_embedding.weight, std=0.01)

        self.input_size += self.embedding_dim

        # Optionally, include charlm embeddings
        self.use_charlm = use_charlm

        if self.use_charlm:
            if charlm_forward_file is None or not os.path.exists(charlm_forward_file):
                raise FileNotFoundError(f'Could not find forward character model: {charlm_forward_file}')
            if charlm_backward_file is None or not os.path.exists(charlm_backward_file):
                raise FileNotFoundError(f'Could not find backward character model: {charlm_backward_file}')
            self.add_unsaved_module('charmodel_forward', CharacterLanguageModel.load(charlm_forward_file, finetune=False))
            self.add_unsaved_module('charmodel_backward', CharacterLanguageModel.load(charlm_backward_file, finetune=False))

            self.input_size += self.charmodel_forward.hidden_dim() + self.charmodel_backward.hidden_dim()

        self.upos_emb_dim = self.model_args["upos_emb_dim"]
        self.upos_to_id = upos_to_id
        if self.upos_emb_dim > 0 and self.upos_to_id is not None:
            # TODO: should leave space for unknown POS?
            self.upos_emb = nn.Embedding(num_embeddings=len(self.upos_to_id),
                                         embedding_dim=self.upos_emb_dim,
                                         padding_idx=0)
            self.input_size += self.upos_emb_dim

        device = next(self.parameters()).device
        # Determine if attn or LSTM should be used
        if self.num_heads > 0:
            self.input_size = utils.round_up_to_multiple(self.input_size, self.num_heads)
            self.multihead_attn = nn.MultiheadAttention(embed_dim=self.input_size, num_heads=self.num_heads, batch_first=True).to(device)
            logger.debug(f"Using attention mechanism with embed dim {self.input_size} and {self.num_heads} attention heads.")
        else:
            self.lstm = nn.LSTM(self.input_size,
                                self.hidden_dim,
                                batch_first=True,
                                bidirectional=True)
            logger.debug(f"Using LSTM mechanism.")

        mlp_input_size = self.hidden_dim * 2 if self.num_heads == 0 else self.input_size
        self.mlp = nn.Sequential(
            nn.Linear(mlp_input_size, 64),
            nn.ReLU(),
            nn.Linear(64, output_dim)
        )

    def get_save_dict(self):
        save_dict = {
            "params": self.state_dict(),
            "label_decoder": self.label_decoder,
            "model_type": self.model_type().name,
            "args": self.model_args,
            "upos_to_id": self.upos_to_id,
            "known_words": self.known_words,
            "target_words": list(self.target_words),
            "target_upos": list(self.target_upos),
        }
        skipped = [k for k in save_dict["params"].keys() if self.is_unsaved_module(k)]
        for k in skipped:
            del save_dict["params"][k]
        return save_dict

    def convert_tags(self, upos_tags: List[List[str]]):
        if self.upos_to_id is not None:
            return [[self.upos_to_id[x] for x in sentence] for sentence in upos_tags]
        return None

    def forward(self, pos_indices: List[int], sentences: List[List[str]], upos_tags: List[List[int]]):
        """
        Computes the forward pass of the neural net

        Args:
            pos_indices (List[int]): A list of the position index of the target token for lemmatization classification in each sentence.
            sentences (List[List[str]]): A list of the token-split sentences of the input data.
            upos_tags (List[List[int]]): A list of the upos tags for each token in every sentence.

        Returns:
            torch.tensor: Output logits of the neural network, where the shape is  (n, output_size) where n is the number of sentences.
        """
        device = next(self.parameters()).device
        batch_size = len(sentences)
        token_ids = []
        delta_token_ids = []
        for words in sentences:
            sentence_token_ids = [self.vocab_map.get(word.lower(), UNK_ID) for word in words]
            sentence_token_ids = torch.tensor(sentence_token_ids, device=device)
            token_ids.append(sentence_token_ids)

            sentence_delta_token_ids = [self.known_word_map.get(word.lower(), 0) for word in words]
            sentence_delta_token_ids = torch.tensor(sentence_delta_token_ids, device=device)
            delta_token_ids.append(sentence_delta_token_ids)

        token_ids = pad_sequence(token_ids, batch_first=True)
        delta_token_ids = pad_sequence(delta_token_ids, batch_first=True)
        embedded = self.embeddings(token_ids) + self.delta_embedding(delta_token_ids)

        if self.upos_emb_dim > 0:
            upos_tags = [torch.tensor(sentence_tags) for sentence_tags in upos_tags]  # convert internal lists to tensors
            upos_tags = pad_sequence(upos_tags, batch_first=True, padding_value=0).to(device)
            pos_emb = self.upos_emb(upos_tags)
            embedded = torch.cat((embedded, pos_emb), 2).to(device)

        if self.use_charlm:
            char_reps_forward = self.charmodel_forward.build_char_representation(sentences)  # takes [[str]]
            char_reps_backward = self.charmodel_backward.build_char_representation(sentences)

            char_reps_forward = pad_sequence(char_reps_forward, batch_first=True)
            char_reps_backward = pad_sequence(char_reps_backward, batch_first=True)

            embedded = torch.cat((embedded, char_reps_forward, char_reps_backward), 2)

        if self.num_heads > 0:

            def positional_encoding(seq_len, d_model, device):
                encoding = torch.zeros(seq_len, d_model, device=device)
                position = torch.arange(0, seq_len, dtype=torch.float, device=device).unsqueeze(1)
                div_term = torch.exp(torch.arange(0, d_model, 2).float() * (-math.log(10000.0) / d_model)).to(device)

                encoding[:, 0::2] = torch.sin(position * div_term)
                encoding[:, 1::2] = torch.cos(position * div_term)

                # Add a new dimension to fit the batch size
                encoding = encoding.unsqueeze(0)
                return encoding

            seq_len, d_model = embedded.shape[1], embedded.shape[2]
            pos_enc = positional_encoding(seq_len, d_model, device=device)

            embedded += pos_enc.expand_as(embedded)

        padded_sequences = pad_sequence(embedded, batch_first=True)
        lengths = torch.tensor([len(seq) for seq in embedded])

        if self.num_heads > 0:
            target_seq_length, src_seq_length = padded_sequences.size(1), padded_sequences.size(1)
            attn_mask = torch.triu(torch.ones(batch_size * self.num_heads, target_seq_length, src_seq_length, dtype=torch.bool), diagonal=1)

            attn_mask = attn_mask.view(batch_size, self.num_heads, target_seq_length, src_seq_length)
            attn_mask = attn_mask.repeat(1, 1, 1, 1).view(batch_size * self.num_heads, target_seq_length, src_seq_length).to(device)

            attn_output, attn_weights = self.multihead_attn(padded_sequences, padded_sequences, padded_sequences, attn_mask=attn_mask)
            # Extract the hidden state at the index of the token to classify
            token_reps = attn_output[torch.arange(attn_output.size(0)), pos_indices]

        else:
            packed_sequences = pack_padded_sequence(padded_sequences, lengths, batch_first=True)
            lstm_out, (hidden, _) = self.lstm(packed_sequences)
            # Extract the hidden state at the index of the token to classify
            unpacked_lstm_outputs, _ = pad_packed_sequence(lstm_out, batch_first=True)
            token_reps = unpacked_lstm_outputs[torch.arange(unpacked_lstm_outputs.size(0)), pos_indices]

        # MLP forward pass
        output = self.mlp(token_reps)
        return output

    def model_type(self):
        return ModelType.LSTM



def find_nth(haystack, needle, n):
    """
    Returns the starting index of the nth occurrence of the substring 'needle' in the string 'haystack'.
    """
    start = haystack.find(needle)
    while start >= 0 and n > 1:
        start = haystack.find(needle, start + len(needle))
        n -= 1
    return start


def round_base(num, base=10):
    """
    Rounding a number to its nearest multiple of the base. round_base(49.2, base=50) = 50.
    """
    return base * round(num / base)


import sys

import json

def max_mwt_length(filenames):
    max_len = 0
    for filename in filenames:
        with open(filename) as f:
            d = json.load(f)
            max_len = max([max_len] + [len(" ".join(x[0][1])) for x in d])
    return max_len

if __name__ == '__main__':
    print(max_max_jlength(sys.argv[1:]))



"""
Utils for training and evaluating multiple models simultaneously
"""

import argparse
import os

from stanza.models.lemma_classifier.train_lstm_model import main as train_lstm_main
from stanza.models.lemma_classifier.train_transformer_model import main as train_tfmr_main
from stanza.models.lemma_classifier.constants import DEFAULT_BATCH_SIZE


change_params_map = {
    "lstm_layer": [16, 32, 64, 128, 256, 512],
    "upos_emb_dim": [5, 10, 20, 30],
    "training_size": [150, 300, 450, 600, 'full'],
}  # TODO: Add attention

def train_n_models(num_models: int, base_path: str, args):

    if args.change_param == "lstm_layer":
        for num_layers in change_params_map.get("lstm_layer", None):
            for i in range(num_models):
                new_save_name = os.path.join(base_path, f"{num_layers}_{i}.pt")
                args.save_name = new_save_name
                args.hidden_dim = num_layers
                train_lstm_main(predefined_args=args)

    if args.change_param == "upos_emb_dim":
        for upos_dim in change_params_map("upos_emb_dim", None):
            for i in range(num_models):
                new_save_name = os.path.join(base_path, f"dim_{upos_dim}_{i}.pt")
                args.save_name = new_save_name
                args.upos_emb_dim = upos_dim
                train_lstm_main(predefined_args=args)

    if args.change_param == "training_size":
        for size in change_params_map.get("training_size", None):
            for i in range(num_models):
                new_save_name = os.path.join(base_path, f"{size}_examples_{i}.pt")
                new_train_file = os.path.join(os.path.dirname(__file__), "data", "processed_ud_en", "combined_train.txt")
                args.save_name = new_save_name
                args.train_file = new_train_file
                train_lstm_main(predefined_args=args)

    if args.change_param == "base":
        for i in range(num_models):
            new_save_name = os.path.join(base_path, f"lstm_model_{i}.pt")
            args.save_name = new_save_name
            args.weighted_loss = False
            train_lstm_main(predefined_args=args)

            if not args.weighted_loss:
                args.weighted_loss = True
                new_save_name = os.path.join(base_path, f"lstm_model_wloss_{i}.pt")
                args.save_name = new_save_name
                train_lstm_main(predefined_args=args)

    if args.change_param == "base_charlm":
        for i in range(num_models):
            new_save_name = os.path.join(base_path, f"lstm_charlm_{i}.pt")
            args.save_name = new_save_name
            train_lstm_main(predefined_args=args)

    if args.change_param == "base_charlm_upos":
        for i in range(num_models):
            new_save_name = os.path.join(base_path, f"lstm_charlm_upos_{i}.pt")
            args.save_name = new_save_name
            train_lstm_main(predefined_args=args)

    if args.change_param == "base_upos":
        for i in range(num_models):
            new_save_name = os.path.join(base_path, f"lstm_upos_{i}.pt")
            args.save_name = new_save_name
            train_lstm_main(predefined_args=args)

    if args.change_param == "attn_model":
        for i in range(num_models):
            new_save_name = os.path.join(base_path, f"attn_model_{args.num_heads}_heads_{i}.pt")
            args.save_name = new_save_name
            train_lstm_main(predefined_args=args)

def train_n_tfmrs(num_models: int, base_path: str, args):

    if args.multi_train_type == "tfmr":

        for i in range(num_models):

            if args.change_param == "bert":
                new_save_name = os.path.join(base_path, f"bert_{i}.pt")
                args.save_name = new_save_name
                args.loss_fn = "ce"
                train_tfmr_main(predefined_args=args)

                new_save_name = os.path.join(base_path, f"bert_wloss_{i}.pt")
                args.save_name = new_save_name
                args.loss_fn = "weighted_bce"
                train_tfmr_main(predefined_args=args)

            elif args.change_param == "roberta":
                new_save_name = os.path.join(base_path, f"roberta_{i}.pt")
                args.save_name = new_save_name
                args.loss_fn = "ce"
                train_tfmr_main(predefined_args=args)

                new_save_name = os.path.join(base_path, f"roberta_wloss_{i}.pt")
                args.save_name = new_save_name
                args.loss_fn = "weighted_bce"
                train_tfmr_main(predefined_args=args)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--hidden_dim", type=int, default=256, help="Size of hidden layer")
    parser.add_argument('--wordvec_pretrain_file', type=str, default=os.path.join(os.path.dirname(__file__), "pretrain", "glove.pt"), help='Exact name of the pretrain file to read')
    parser.add_argument("--charlm", action='store_true', dest='use_charlm', default=False, help="Whether not to use the charlm embeddings")
    parser.add_argument('--charlm_shorthand', type=str, default=None, help="Shorthand for character-level language model training corpus.")
    parser.add_argument("--charlm_forward_file", type=str, default=os.path.join(os.path.dirname(__file__), "charlm_files", "1billion_forward.pt"), help="Path to forward charlm file")
    parser.add_argument("--charlm_backward_file", type=str, default=os.path.join(os.path.dirname(__file__), "charlm_files", "1billion_backwards.pt"), help="Path to backward charlm file")
    parser.add_argument("--upos_emb_dim", type=int, default=20, help="Dimension size for UPOS tag embeddings.")
    parser.add_argument("--use_attn", action='store_true', dest='attn', default=False, help='Whether to use multihead attention instead of LSTM.')
    parser.add_argument("--num_heads", type=int, default=0, help="Number of heads to use for multihead attention.")
    parser.add_argument("--save_name", type=str, default=os.path.join(os.path.dirname(__file__), "saved_models", "lemma_classifier_model_weighted_loss_charlm_new.pt"), help="Path to model save file")
    parser.add_argument("--lr", type=float, default=0.001, help="learning rate")
    parser.add_argument("--num_epochs", type=float, default=10, help="Number of training epochs")
    parser.add_argument("--batch_size", type=int, default=DEFAULT_BATCH_SIZE, help="Number of examples to include in each batch")
    parser.add_argument("--train_file", type=str, default=os.path.join(os.path.dirname(__file__), "data", "processed_ud_en", "combined_train.txt"), help="Full path to training file")
    parser.add_argument("--weighted_loss", action='store_true', dest='weighted_loss', default=False, help="Whether to use weighted loss during training.")
    parser.add_argument("--eval_file", type=str, default=os.path.join(os.path.dirname(__file__), "data", "processed_ud_en", "combined_dev.txt"), help="Path to dev file used to evaluate model for saves")
    # Tfmr-specific args
    parser.add_argument("--model_type", type=str, default="roberta", help="Which transformer to use ('bert' or 'roberta')")
    parser.add_argument("--bert_model", type=str, default=None, help="Use a specific transformer instead of the default bert/roberta")
    parser.add_argument("--loss_fn", type=str, default="weighted_bce", help="Which loss function to train with (e.g. 'ce' or 'weighted_bce')")
    # Multi-model train args
    parser.add_argument("--multi_train_type", type=str, default="lstm", help="Whether you are attempting to multi-train an LSTM or transformer")
    parser.add_argument("--multi_train_count", type=int, default=5, help="Number of each model to build")
    parser.add_argument("--base_path", type=str, default=None, help="Path to start generating model type for.")
    parser.add_argument("--change_param", type=str, default=None, help="Which hyperparameter to change when training")


    args = parser.parse_args()

    if args.multi_train_type == "lstm":
        train_n_models(num_models=args.multi_train_count,
                       base_path=args.base_path,
                       args=args)
    elif args.multi_train_type == "tfmr":
        train_n_tfmrs(num_models=args.multi_train_count,
                      base_path=args.base_path,
                      args=args)
    else:
        raise ValueError(f"Improper input {args.multi_train_type}")

if __name__ == "__main__":
    main()
